"""Module to load pickled feature dictionaries."""

import os
import pickle

MODULE_PATH = os.path.abspath(os.path.dirname(__file__))
PKL_BASE_PATH = os.path.join(MODULE_PATH, "pickled_feature_dicts")


def load_pickled_feature(pickle_filename: str):
    """
    Load a pickled feature dictionary.

    Parameters
    ----------
    pickle_filename : str
        Name of the pickled file.

    Returns
    -------
    dict
        Loaded pickled dictionary.
    """

    with open(os.path.join(PKL_BASE_PATH, pickle_filename), "rb") as f:
        return pickle.load(f)


PTM_LOSS_LOOKUP = load_pickled_feature("saved_loss_atoms.pkl")
PTM_MOD_DELTA_MASS_LOOKUP = load_pickled_feature("mz_diff.pkl")
PTM_GAIN_LOOKUP = load_pickled_feature("saved_gained_atoms.pkl")
PTM_ATOM_COUNT_LOOKUP = load_pickled_feature("saved_ac_count.pkl")
PTM_RED_SMILES_LOOKUP = load_pickled_feature("red_smiles.pkl")



import torch.nn as nn


class BiGRUSequentialEncoder(nn.Module):
    """Encoder class needed to handle two GRU outputs in torch.

    No implementation using nn.Sequential() possible.

    Args:
        embedding_output_dim (int): The size of the embedding output dimension. Defaults to 16.
        recurrent_layers_sizes (tuple): The sizes of the recurrent layers. Defaults to (256, 512).
        dropout_rate (float): The dropout rate used in the encoder layers. Defaults to 0.5.
    """

    def __init__(
        self,
        embedding_output_dim,
        recurrent_layers_sizes,
        dropout_rate,
    ):
        super(BiGRUSequentialEncoder, self).__init__()

        self.bidirectional_GRU = nn.GRU(
            input_size=embedding_output_dim,
            hidden_size=recurrent_layers_sizes[0],
            batch_first=True,
            bidirectional=True,
        )
        self.encoder_dropout1 = nn.Dropout(dropout_rate)
        self.unidirectional_GRU = nn.GRU(
            input_size=recurrent_layers_sizes[0] * 2,
            hidden_size=recurrent_layers_sizes[1],
            batch_first=True,
            bidirectional=False,
        )
        self.encoder_dropout2 = nn.Dropout(dropout_rate)

    def forward(self, inputs):
        x, _ = self.bidirectional_GRU(inputs)
        x = self.encoder_dropout1(x)
        x, _ = self.unidirectional_GRU(x)
        x = self.encoder_dropout2(x)
        return x



from ..config import _BACKEND, PYTORCH_BACKEND, TENSORFLOW_BACKEND

if _BACKEND in TENSORFLOW_BACKEND:
    from .chargestate import adjusted_mean_absolute_error, adjusted_mean_squared_error
    from .rt_eval import TimeDeltaMetric, timedelta
elif _BACKEND in PYTORCH_BACKEND:
    from .chargestate_torch import (
        adjusted_mean_absolute_error,
        adjusted_mean_squared_error,
    )
    from .rt_eval_torch import TimeDeltaMetric, timedelta

__all__ = [
    "adjusted_mean_absolute_error",
    "adjusted_mean_squared_error",
    "timedelta",
    "TimeDeltaMetric",
]



from .DetectabilityReport import DetectabilityReport
from .IntensityReport import IntensityReport
from .RetentionTimeReport import RetentionTimeReport

__all__ = [
    "RetentionTimeReport",
    "IntensityReport",
    "DetectabilityReport",
]



import warnings

import tensorflow as tf

from ..constants import ALPHABET_UNMOD
from ..layers.attention import AttentionLayer

"""
This module contains a deep learning model for precursor charge state prediction, inspired by Prosit's architecture.
The model is provided in three flavours of predicting precursor charge states:

1. Dominant Charge State Prediction:
   - Task: Predict the dominant charge state of a given peptide sequence.
   - Model: Uses a multi-class classification approach to predict the most likely charge state.

2. Observed Charge State Prediction:
   - Task: Predict the observed charge states for a given peptide sequence.
   - Model: Uses a multi-label classification approach to predict all possible charge states.

3. Relative Charge State Prediction:
   - Task: Predict the proportion of each charge state for a given peptide sequence.
   - Model: Uses a regression approach to predict the proportion of each charge state.
"""


@tf.keras.utils.register_keras_serializable(package="dlomix")
class ChargeStatePredictor(tf.keras.Model):
    """
    Precursor Charge State Prediction Model for predicting either:
    * the dominant charge state or
    * all observed charge states or
    * the relative charge state distribution
    of a peptide sequence.

    Args:
        embedding_output_dim (int): The size of the embedding output dimension. Defaults to 16.
        seq_length (int): The length of the input sequence. Defaults to 30.
        alphabet (dict): Dictionary mapping for the alphabet (the amino acids in this case). Defaults to ALPHABET_UNMOD.
        dropout_rate (float): The dropout rate used in the encoder layers. Defaults to 0.5.
        latent_dropout_rate (float): The dropout rate for the latent space. Defaults to 0.1.
        recurrent_layers_sizes (tuple): The sizes of the recurrent layers. Defaults to (256, 512).
        regressor_layer_size (int): The size of the regressor layer. Defaults to 512.
        num_classes (int): The number of classes for the output corresponding to charge states available in the data. Defaults to 6.
        model_flavour (str): The type of precursor charge state prediction to be done.
            Can be either "dominant", "observed" or "relative".
            Defaults to "relative".
    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        alphabet=ALPHABET_UNMOD,
        dropout_rate=0.5,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
        num_classes=6,
        model_flavour="relative",
    ):
        super(ChargeStatePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(alphabet) + 1

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes

        if model_flavour == "relative":
            # regression problem
            self.final_activation = "linear"
        elif model_flavour == "observed":
            # multi-label multi-class classification problem
            self.final_activation = "sigmoid"
        elif model_flavour == "dominant":
            # multi-class classification problem
            self.final_activation = "softmax"
        else:
            warnings.warn(f"{model_flavour} not available")
            exit

        self.embedding = tf.keras.layers.Embedding(
            input_dim=self.embeddings_count,
            output_dim=embedding_output_dim,
            input_length=seq_length,
        )
        self._build_encoder()

        self.attention = AttentionLayer()

        self.regressor = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(self.regressor_layer_size, activation="relu"),
                tf.keras.layers.Dropout(rate=self.latent_dropout_rate),
            ]
        )

        self.output_layer = tf.keras.layers.Dense(
            num_classes, activation=self.final_activation
        )

    def _build_encoder(self):
        self.encoder = tf.keras.Sequential(
            [
                tf.keras.layers.Bidirectional(
                    tf.keras.layers.GRU(
                        units=self.recurrent_layers_sizes[0], return_sequences=True
                    )
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
                tf.keras.layers.GRU(
                    units=self.recurrent_layers_sizes[1], return_sequences=True
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
            ]
        )

    def call(self, inputs):
        x = self.embedding(inputs)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        return x



import logging
from collections import OrderedDict
from collections.abc import Sequence

import torch
import torch.nn as nn

from ..constants import ALPHABET_UNMOD
from ..data.processing.feature_extractors import FEATURE_EXTRACTORS_PARAMETERS
from ..layers.attention_torch import AttentionLayer
from ..layers.bi_gru_seq_encoder_torch import BiGRUSequentialEncoder
from ..layers.gru_seq_decoder_torch import GRUSequentialDecoder

logger = logging.getLogger("dlomix.models.prosit_torch")


class PrositRetentionTimePredictor(nn.Module):
    """
    Implementation of the Prosit model for retention time prediction.

    Parameters
    -----------
    embedding_output_dim : int, optional
        Size of the embeddings to use. Defaults to 16.
    seq_length : int, optional
        Sequence length of the peptide sequences. Defaults to 30.
    alphabet : dict, optional
        Dictionary mapping for the alphabet (the amino acids in this case). Defaults to ALPHABET_UNMOD.
    dropout_rate : float, optional
        Probability to use for dropout layers in the encoder. Defaults to 0.5.
    latent_dropout_rate : float, optional
        Probability to use for dropout layers in the regressor layers after encoding. Defaults to 0.1.
    recurrent_layers_sizes : tuple, optional
        A tuple of 2 values for the sizes of the two GRU layers in the encoder. Defaults to (256, 512).
    regressor_layer_size : int, optional
        Size of the dense layer in the regressor after the encoder. Defaults to 512.


    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        alphabet=ALPHABET_UNMOD,
        dropout_rate=0.5,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
    ):
        super(PrositRetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(alphabet)

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes
        self.embedding_output_dim = embedding_output_dim
        self.seq_length = seq_length

        self.embedding = nn.Embedding(
            num_embeddings=self.embeddings_count,
            embedding_dim=embedding_output_dim,
            padding_idx=0,  # TODO check this
        )

        self.encoder = BiGRUSequentialEncoder(
            embedding_output_dim, self.recurrent_layers_sizes, self.dropout_rate
        )

        self.attention = AttentionLayer(
            feature_dim=self.recurrent_layers_sizes[1], seq_len=self.seq_length
        )

        self.regressor = nn.Sequential(
            OrderedDict(
                [
                    (
                        "dense",
                        nn.Linear(
                            in_features=self.recurrent_layers_sizes[1],
                            out_features=self.regressor_layer_size,
                        ),
                    ),
                    ("activation_relu", nn.ReLU()),
                    ("regressor_dropout", nn.Dropout(self.latent_dropout_rate)),
                ]
            )
        )

        self.output_layer = nn.Linear(
            in_features=self.regressor_layer_size, out_features=1
        )

    def forward(self, inputs, **kwargs):
        x = self.embedding(inputs)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        return x


class PrositIntensityPredictor(nn.Module):
    """
    Prosit model for intensity prediction with configurable branches for PTM features and metadata.

    Parameters
    ----------
    input_keys : dict, optional
        Dictionary mapping for the input keys to look for in the input dict. Defaults to None, which uses default required keys only "seqeuence".
    meta_data_keys : list or dict, optional
        List or dict of keys for the meta data inputs to use. Defaults to None (no meta data).
    alphabet : dict, optional
        Dictionary mapping for the alphabet (the amino acids in this case). Defaults to ALPHABET_UNMOD.
    with_termini : bool, optional
        Whether to include terminal tokens in the sequence embedding. Defaults to False.
    embedding_output_dim : int, optional
        Size of the embeddings to use. Defaults to 16.
    seq_length : int, optional
        Sequence length of the peptide sequences. Defaults to 30.
    len_fragment_ion : int, optional
        Number of fragment ions to predict. Defaults to 6.
    dropout_rate : float, optional
        Probability to use for dropout layers in the encoder. Defaults to 0.2.
    latent_dropout_rate : float, optional
        Probability to use for dropout layers in the regressor layers after encoding. Defaults to 0.1.
    recurrent_layers_sizes : tuple, optional
        A tuple of 2 values for the sizes of the two GRU layers in the encoder. Defaults to (256, 512).
    regressor_layer_size : int, optional
        Size of the dense layer in the regressor after the encoder. Defaults to 512.
    use_meta_data : bool, optional
        Whether to use meta data inputs. Defaults to False.
    use_prosit_ptm_features : bool, optional
        Whether to use Prosit PTM features as input. Defaults to False.
    use_instrument_embedding : bool, optional
        Whether to use instrument type embedding as part of the meta data. Defaults to False.
    instrument_input_dim : int, optional
        Number of unique instrument types for embedding. Defaults to 3.
    instrument_output_dim : int, optional
        Size of the instrument type embedding. Defaults to 2.



    Attributes
    ----------

    REQUIRED_INPUT_SEQUENCE_KEY : str
        Key for the required sequence input in the input dictionary.
    DEFAULT_INPUT_KEYS : dict
        Default mapping of input keys for various inputs.
    META_DATA_KEYS : list
        List of possible meta data keys that can be used.
    PTM_INPUT_KEYS : list
        List of keys for the PTM feature inputs. See `dlomix.data.processing.feature_extractors.FEATURE_EXTRACTORS_PARAMETERS` for details.


    Notes
    -----
    This model is a flexible implementation of the Prosit intensity predictor, allowing for optional
    inclusion of PTM features and meta data inputs. The model architecture consists of embedding layers,
    bidirectional GRU encoders, attention mechanisms, and dense regressor layers.






    """

    REQUIRED_INPUT_SEQUENCE_KEY = "SEQUENCE_KEY"

    DEFAULT_INPUT_KEYS = {
        REQUIRED_INPUT_SEQUENCE_KEY: "sequence",
        "COLLISION_ENERGY_KEY": "collision_energy",
        "PRECURSOR_CHARGE_KEY": "precursor_charge",
        "FRAGMENTATION_TYPE_KEY": "fragmentation_type",
        "INSTRUMENT_TYPE_KEY": "instrument_type",
    }

    # can be extended to include all possible meta data
    META_DATA_KEYS = [
        "COLLISION_ENERGY_KEY",
        "PRECURSOR_CHARGE_KEY",
        "FRAGMENTATION_TYPE_KEY",
        "INSTRUMENT_TYPE_KEY",
    ]

    # retrieve the Lookup PTM feature keys
    PTM_INPUT_KEYS = [*FEATURE_EXTRACTORS_PARAMETERS.keys()]

    def __init__(
        self,
        input_keys=None,
        meta_data_keys=None,
        alphabet=None,
        with_termini=False,
        embedding_output_dim=16,
        seq_length=30,
        len_fragment_ion=6,
        dropout_rate=0.2,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
        use_meta_data=False,
        use_prosit_ptm_features=False,
        use_instrument_embedding=False,
        instrument_input_dim=3,
        instrument_output_dim=2,
        **kwargs,
    ):
        super(PrositIntensityPredictor, self).__init__(**kwargs)

        # Store all configuration parameters
        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = tuple(recurrent_layers_sizes)
        self.embedding_output_dim = embedding_output_dim
        self.raw_seq_length = seq_length
        self.len_fragment_ion = len_fragment_ion
        self.use_prosit_ptm_features = use_prosit_ptm_features
        self.with_termini = with_termini
        self.use_meta_data = use_meta_data
        self.use_instrument_embedding = use_instrument_embedding
        self.instrument_input_dim = instrument_input_dim
        self.instrument_output_dim = instrument_output_dim

        # handle default and fallback attributes
        self._handle_alphabet_and_keys(alphabet, input_keys, meta_data_keys)

        self._validate_config()

        self._compute_attributes()

        # Build layers
        self._build_embedding_layers()
        self._build_encoders()
        self._build_decoder()
        self.attention = AttentionLayer(
            feature_dim=regressor_layer_size, seq_len=seq_length
        )
        self._build_meta_data_fusion_layer()
        self._build_regressor()

    def _handle_alphabet_and_keys(self, alphabet, input_keys, meta_data_keys):
        # Handle alphabet
        if alphabet is not None:
            self.alphabet = dict(alphabet)
        else:
            self.alphabet = dict(ALPHABET_UNMOD)  # fallback to unmodified amino acids

        # Handle input_keys
        if input_keys:
            self.input_keys = dict(input_keys)
        else:
            self.input_keys = {
                self.REQUIRED_INPUT_SEQUENCE_KEY: "sequence"
            }  # minimal required key for the sequence input

        # Handle meta_data_keys
        if not meta_data_keys:
            self.meta_data_keys = []
        else:
            # handle dict in case user passed a mapping, take the values in a list to use for lookup
            if isinstance(meta_data_keys, dict):
                self.meta_data_keys = list(meta_data_keys.values())
            elif isinstance(meta_data_keys, list):
                self.meta_data_keys = list(meta_data_keys)
            else:
                raise ValueError(
                    "meta_data_keys should be either a list of strings or a dict mapping. Provided type: "
                    f"{type(meta_data_keys)}, and value: {meta_data_keys}"
                )

    def _validate_config(self):
        if self.use_meta_data and not self.meta_data_keys:
            raise ValueError(
                "use_meta_data=True requires meta_data_keys to be provided as a list of keys."
            )

        if (
            self.use_instrument_embedding
            and "INSTRUMENT_TYPE_KEY" not in self.input_keys
        ):
            raise ValueError(
                "use_instrument_embedding=True requires 'INSTRUMENT_TYPE_KEY' in input_keys"
            )

    def _compute_attributes(self):
        # Compute derived attributes (will be recomputed during deserialization)
        self.max_ion = self.raw_seq_length - 1

        self.seq_length = (
            self.raw_seq_length + 2 if self.with_termini else self.raw_seq_length
        )

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(self.alphabet)

    def _build_embedding_layers(self):
        self.embedding = nn.Embedding(
            num_embeddings=self.embeddings_count,
            embedding_dim=self.embedding_output_dim,
        )

        self.instrument_embedding = None
        if self.use_instrument_embedding:
            self.instrument_embedding = nn.Embedding(
                num_embeddings=self.instrument_input_dim,
                embedding_dim=self.instrument_output_dim,
            )

    def _build_meta_data_fusion_layer(self):
        if self.meta_data_keys:
            self.meta_data_fusion_layer = MetaDataFusionBlock(max_ion=self.max_ion)

    def _build_encoders(self):
        # sequence encoder -> always present
        gru_features_input_size = self.embedding_output_dim

        if self.use_prosit_ptm_features:
            gru_features_input_size = self.embedding_output_dim * 2

        self.sequence_encoder = BiGRUSequentialEncoder(
            embedding_output_dim=gru_features_input_size,
            recurrent_layers_sizes=self.recurrent_layers_sizes,
            dropout_rate=self.dropout_rate,
        )

        # meta data encoder -> optional, only if meta data keys are provided
        self.meta_encoder = None
        if self.use_meta_data:
            self.meta_encoder = nn.Sequential(
                OrderedDict(
                    [
                        (
                            "meta_dense",
                            nn.LazyLinear(out_features=self.recurrent_layers_sizes[1]),
                        ),
                        ("dropout", nn.Dropout(p=self.dropout_rate)),
                    ]
                )
            )

        # ptm encoder -> optional, only if ptm flag is provided
        self.ptm_input_encoder, self.ptm_aa_fusion = None, None
        if self.use_prosit_ptm_features:
            self.ptm_input_encoder = nn.Sequential(
                Concatenate(dim=-1),
                nn.LazyLinear(out_features=self.regressor_layer_size // 2),
                nn.Dropout(p=self.dropout_rate),
                nn.LazyLinear(out_features=self.embedding_output_dim * 4),
                nn.Dropout(p=self.dropout_rate),
                nn.LazyLinear(out_features=self.embedding_output_dim),
                nn.Dropout(p=self.dropout_rate),
            )

            self.ptm_aa_fusion = Concatenate(dim=-1)

    def _build_decoder(self):
        self.decoder = GRUSequentialDecoder(
            recurrent_layers_sizes=self.recurrent_layers_sizes,
            dropout_rate=self.dropout_rate,
            max_ion=self.max_ion,
        )

    def _build_regressor(self):
        self.regressor = nn.Sequential(
            OrderedDict(
                [
                    ("time_dense", nn.LazyLinear(out_features=self.len_fragment_ion)),
                    ("activation", nn.LeakyReLU()),
                    ("output", nn.Flatten()),
                ]
            )
        )

    def forward(self, inputs, **kwargs):
        # Handle dict input, complex case: multiple inputs
        if isinstance(inputs, dict):
            return self._forward_dict(inputs, **kwargs)

        # Handle single input, simple case: sequence only
        peptides_in = inputs
        return self._forward_sequence_only(peptides_in, **kwargs)

    def _forward_sequence_only(self, peptides_in, **kwargs):
        x = self.embedding(peptides_in)
        x = self.sequence_encoder(x)
        x = self.attention(x)
        x = torch.unsqueeze(x, dim=1)
        x = self.decoder(x)
        x = self.regressor(x)
        return x

    def _forward_dict(self, inputs, **kwargs):
        missing_input_keys = [k for k in self.input_keys.values() if k not in inputs]
        if missing_input_keys:
            raise ValueError(f"Missing required input keys: {missing_input_keys}")

        meta_data = []
        encoded_meta = None
        encoded_ptm = None

        # collect instrument embedding if enabled and add to meta data
        if self.use_instrument_embedding:
            instrument_type = inputs.get(self.input_keys["INSTRUMENT_TYPE_KEY"])
            instrument_embedded = self.instrument_embedding(instrument_type)
            meta_data.append(instrument_embedded)

        # collect meta data from the input dict
        if self.use_meta_data:
            missing_meta_keys = [k for k in self.meta_data_keys if k not in inputs]
            if missing_meta_keys:
                raise ValueError(
                    f"Missing required metadata inputs: {missing_meta_keys}"
                )

            meta_data.extend(
                self._collect_values_from_inputs_if_exists(inputs, self.meta_data_keys)
            )

        # collect PTM features from the input dict
        if self.use_prosit_ptm_features:
            ptm_keys_exist = [k for k in self.PTM_INPUT_KEYS if k in inputs]
            if not ptm_keys_exist:
                raise ValueError(
                    f"At least one PTM input feature is required when use_prosit_ptm_features=True. Missing all of: {self.PTM_INPUT_KEYS}"
                )

            ptm_ac_features = self._collect_values_from_inputs_if_exists(
                inputs, PrositIntensityPredictor.PTM_INPUT_KEYS
            )

        peptides_in = inputs[self.input_keys[self.REQUIRED_INPUT_SEQUENCE_KEY]]

        x = self.embedding(peptides_in)

        # encode and fuse PTM features (before going into the GRU sequence encoder)
        if self.use_prosit_ptm_features:
            encoded_ptm = self.ptm_input_encoder(ptm_ac_features)
            x = self.ptm_aa_fusion([x, encoded_ptm])

        x = self.sequence_encoder(x)
        x = self.attention(x)

        if self.use_meta_data:
            if isinstance(meta_data, list):
                meta_data = torch.cat(meta_data, dim=-1)
            encoded_meta = self.meta_encoder(meta_data)
            x = self.meta_data_fusion_layer([x, encoded_meta])
        else:
            # no metadata -> add a dimension to comply with the shape
            x = torch.unsqueeze(x, dim=1)

        x = self.decoder(x)
        x = self.regressor(x)

        return x

    def _collect_values_from_inputs_if_exists(self, inputs, keys_mapping):
        collected_values = []
        keys = []

        if isinstance(keys_mapping, dict):
            keys = list(keys_mapping.values())

        elif isinstance(keys_mapping, Sequence):
            keys = list(keys_mapping)

        for key_in_inputs in keys:
            # get the input under the specified key if exists
            single_input = inputs.get(key_in_inputs, None)
            if single_input is not None:
                if single_input.ndim == 1:
                    single_input = torch.unsqueeze(single_input, dim=-1)
                collected_values.append(single_input)
        return collected_values


class Concatenate(nn.Module):
    def __init__(self, dim=-1):
        super(Concatenate, self).__init__()
        self.dim = dim

    def forward(self, inputs):
        for x in inputs:
            logger.debug("concatenating: %s ", x.shape)
        return torch.cat(inputs, dim=self.dim)


class MetaDataFusionBlock(torch.nn.Module):
    def __init__(self, max_ion):
        super(MetaDataFusionBlock, self).__init__()
        self.max_ion = max_ion

    def forward(self, x):
        #  x is a tuple of (features, metadata)
        features, metadata = x
        # Multiply operation
        multiplied = features * metadata
        # RepeatVector equivalent - expand along dimension to repeat max_ion times
        repeated = multiplied.unsqueeze(1).expand(-1, self.max_ion, -1)
        return repeated



from typing import Callable, Dict, List, Optional, Union

from .dataset import PeptideDataset
from .dataset_config import DatasetConfig
from .dataset_utils import EncodingScheme


class FragmentIonIntensityDataset(PeptideDataset):
    """
    A dataset class for handling fragment ion intensity data.

    This class inherits from PeptideDataset and uses DatasetConfig to manage
    its configuration. All parameters are passed to DatasetConfig and used
    throughout the parent class implementation.

    Args:
        data_source (Optional[Union[str, List]]): The path or list of paths to the data source file(s).
        val_data_source (Optional[Union[str, List]]): The path or list of paths to the validation data source file(s).
        test_data_source (Optional[Union[str, List]]): The path or list of paths to the test data source file(s).
        data_format (str): The format of the data source file(s). Default is "parquet".
        sequence_column (str): The name of the column containing the peptide sequences. Default is "modified_sequence".
        label_column (str): The name of the column containing the intensity labels. Default is "intensities_raw".
        val_ratio (float): The ratio of validation data to split from the training data. Default is 0.2.
        max_seq_len (Union[int, str]): The maximum length of the peptide sequences. Default is 30.
        dataset_type (str): The type of dataset to use (e.g., "tf" for TensorFlow dataset). Default is "tf". Fallback is to TensorFlow dataset tensors.
        batch_size (int): The batch size for training and evaluation. Default is 64.
        shuffle (bool): Whether to shuffle the data. Default is False.
        model_features (Optional[List[str]]): The list of features to use for the model. Default is None.
        dataset_columns_to_keep (Optional[List[str]]): The list of columns to keep in the dataset. Default is None.
        features_to_extract (Optional[List[Union[Callable, str]]]): The list of features to extract from the dataset. Default is None.
        pad (bool): Whether to pad the sequences to the maximum length. Default is True.
        padding_value (str): The value to use for padding. Default is '-'.
        alphabet (Optional[Dict]): The mapping of characters to integers for encoding the sequences. Default is None to trigger learning the alphabet.
        with_termini (bool): Whether to add the N- and C-termini in the sequence column, even if they do not exist. Defaults to True.
        encoding_scheme (Union[str, EncodingScheme]): The encoding scheme to use for encoding the sequences. Default is EncodingScheme.UNMOD.
        processed (bool): Whether the data has been preprocessed before or not. Default is False.
        enable_tf_dataset_cache (bool): Flag to indicate whether to enable TensorFlow Dataset caching (call `.cache()` on the generated TF Datasets). Default is False.
        disable_cache (bool): Whether to disable Hugging Face datasets caching. Default is False.
        auto_cleanup_cache (bool): Whether to automatically clean up the cache. Default is True.
        num_proc (Optional[int]): Number of processes to use for dataset processing. Use -1 for all available processors, None for single-process mode, or a positive integer. Default is -1.
        batch_processing_size (int): Size of batches for processing. Default is 1000.
        torch_dataloader_kwargs (Optional[Dict]): Additional keyword arguments to pass to PyTorch DataLoader. Default is None.
        **kwargs: Additional arguments to pass to the parent class.
    """

    def __init__(
        self,
        data_source: Optional[Union[str, List]] = None,
        val_data_source: Optional[Union[str, List]] = None,
        test_data_source: Optional[Union[str, List]] = None,
        data_format: str = "parquet",
        sequence_column: str = "modified_sequence",
        label_column: str = "intensities_raw",
        val_ratio: float = 0.2,
        max_seq_len: Union[int, str] = 30,
        dataset_type: str = "tf",
        batch_size: int = 64,
        shuffle: bool = False,
        model_features: Optional[List[str]] = None,
        dataset_columns_to_keep: Optional[List[str]] = None,
        features_to_extract: Optional[List[Union[Callable, str]]] = None,
        pad: bool = True,
        padding_value: str = "-",
        alphabet: Optional[Dict] = None,
        with_termini: bool = True,
        encoding_scheme: Union[str, EncodingScheme] = EncodingScheme.UNMOD,
        processed: bool = False,
        enable_tf_dataset_cache: bool = False,
        disable_cache: bool = False,
        auto_cleanup_cache: bool = True,
        num_proc: Optional[int] = -1,
        batch_processing_size: int = 1000,
        torch_dataloader_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        # Create config kwargs dictionary from all local parameters
        # All parameters are used by DatasetConfig even though they're marked as unused in this method
        config_kwargs = {
            k: v
            for k, v in locals().items()
            if k not in ["self", "__class__", "kwargs"]
        }
        super().__init__(DatasetConfig(**config_kwargs), **kwargs)



import json
import os
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Union

from .dataset_utils import EncodingScheme, validate_num_proc_value


@dataclass
class DatasetConfig:
    """
    Configuration class for the dataset.
    """

    data_source: Union[str, List]
    val_data_source: Union[str, List]
    test_data_source: Union[str, List]
    data_format: str
    sequence_column: str
    label_column: List[str]
    val_ratio: float
    max_seq_len: int
    dataset_type: str
    batch_size: int
    shuffle: bool
    model_features: List[str]
    dataset_columns_to_keep: Optional[List[str]]
    features_to_extract: Optional[List[Union[Callable, str]]]
    pad: bool
    padding_value: str
    alphabet: Dict
    with_termini: bool
    encoding_scheme: Union[str, EncodingScheme]
    processed: bool
    enable_tf_dataset_cache: bool
    disable_cache: bool
    auto_cleanup_cache: bool
    num_proc: Optional[int]
    batch_processing_size: int
    torch_dataloader_kwargs: Optional[Dict] = field(default_factory=dict)

    # validate input parameters
    def __post_init__(self):
        # sequence length validation
        if self.max_seq_len <= 0:
            raise ValueError(
                f"Max sequence length provided is an integer but not a valid value: {self.max_seq_len}, only positive non-zero values are allowed."
            )

        # label column validation, either a string or a list of strings
        if not isinstance(self.label_column, (str, list)):
            raise ValueError(
                "The label_column parameter should be a string or a list of strings."
            )
        elif isinstance(self.label_column, str):
            self.label_column = [self.label_column]

        validate_num_proc_value(self.num_proc)

    def save_config_json(self, path: str):
        """
        Save the configuration to a json file.

        Args:
            path (str): Path to the json file.
        """

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.__dict__, f, default=repr)

    @staticmethod
    def load_config_json(path: str):
        """
        Load the configuration from a json file.

        Args:
            path (str): Path to the json file.

        Returns:
            DatasetConfig: The configuration object.
        """

        with open(path, "r", encoding="utf-8") as f:
            config = json.load(f)
        return DatasetConfig(**config)



import warnings
from collections import OrderedDict

import torch.nn as nn

from ..constants import ALPHABET_UNMOD
from ..layers.attention_torch import AttentionLayer
from ..layers.bi_gru_seq_encoder_torch import BiGRUSequentialEncoder

"""
This module contains a deep learning model for precursor charge state prediction, inspired by Prosit's architecture.
The model is provided in three flavours of predicting precursor charge states:

1. Dominant Charge State Prediction:
   - Task: Predict the dominant charge state of a given peptide sequence.
   - Model: Uses a multi-class classification approach to predict the most likely charge state.

2. Observed Charge State Prediction:
   - Task: Predict the observed charge states for a given peptide sequence.
   - Model: Uses a multi-label classification approach to predict all possible charge states.

3. Relative Charge State Prediction:
   - Task: Predict the proportion of each charge state for a given peptide sequence.
   - Model: Uses a regression approach to predict the proportion of each charge state.
"""


class ChargeStatePredictor(nn.Module):
    """
    Precursor Charge State Prediction Model for predicting either:
    * the dominant charge state or
    * all observed charge states or
    * the relative charge state distribution
    of a peptide sequence.

    batch_first used internally (batch, sequence, feature)

    Args:
        embedding_output_dim (int): The size of the embedding output dimension. Defaults to 16.
        seq_length (int): The length of the input sequence. Defaults to 30.
        alphabet (dict): Dictionary mapping for the alphabet (the amino acids in this case). Defaults to ALPHABET_UNMOD.
        dropout_rate (float): The dropout rate used in the encoder layers. Defaults to 0.5.
        latent_dropout_rate (float): The dropout rate for the latent space. Defaults to 0.1.
        recurrent_layers_sizes (tuple): The sizes of the recurrent layers. Defaults to (256, 512).
        regressor_layer_size (int): The size of the regressor layer. Defaults to 512.
        num_classes (int): The number of classes for the output corresponding to charge states available in the data. Defaults to 6.
        model_flavour (str): The type of precursor charge state prediction to be done.
            Can be either "dominant", "observed" or "relative". Defaults to "relative".
    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        alphabet=ALPHABET_UNMOD,
        dropout_rate=0.5,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
        num_classes=6,
        model_flavour="relative",
    ):
        super(ChargeStatePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(alphabet) + 1
        self.seq_length = seq_length

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes

        if model_flavour == "relative":
            # regression problem
            self.final_activation = nn.Identity()  # == "linear activation" in torch
        elif model_flavour == "observed":
            # multi-label multi-class classification problem
            self.final_activation = nn.Sigmoid()
        elif model_flavour == "dominant":
            # multi-class classification problem
            self.final_activation = nn.Identity()
            # in contrast to tf, don't use Softmax here, cause already included in CrossEntropyLoss, which is to be used for dominant case
        else:
            warnings.warn(f"{model_flavour} not available")
            exit

        self.embedding = nn.Embedding(
            num_embeddings=self.embeddings_count,
            embedding_dim=embedding_output_dim,
            padding_idx=0,
        )

        self.encoder = BiGRUSequentialEncoder(
            embedding_output_dim, self.recurrent_layers_sizes, self.dropout_rate
        )

        self.attention = AttentionLayer(
            feature_dim=self.recurrent_layers_sizes[1], seq_len=self.seq_length
        )

        self.regressor = nn.Sequential(
            OrderedDict(
                [
                    (
                        "dense",
                        nn.Linear(
                            in_features=self.recurrent_layers_sizes[1],
                            out_features=self.regressor_layer_size,
                        ),
                    ),
                    ("activation_relu", nn.ReLU()),
                    ("regressor_dropout", nn.Dropout(self.latent_dropout_rate)),
                ]
            )
        )

        self.output_layer = nn.Linear(
            in_features=self.regressor_layer_size, out_features=num_classes
        )

        self.activation = self.final_activation

    def forward(self, inputs):
        """
        Forward pass of the model.

        Parameters
        ----------
        x : tensor
            Input tensor (shape: [batch_size, seq_length]).

        Returns
        -------
        tensor
            Predicted output (shape: [batch_size, num_classes]).
        """
        x = self.embedding(inputs)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        x = self.activation(x)
        return x



from ..config import _BACKEND, PYTORCH_BACKEND, TENSORFLOW_BACKEND

__all__ = []

if _BACKEND in TENSORFLOW_BACKEND:
    from .intensity import masked_pearson_correlation_distance, masked_spectral_distance

elif _BACKEND in PYTORCH_BACKEND:
    from .intensity_torch import (
        masked_pearson_correlation_distance,
        masked_spectral_distance,
    )
    from .ionmob_torch import MaskedIonmobLoss

    __all__.append(
        "MaskedIonmobLoss",
    )

__all__.extend(["masked_pearson_correlation_distance", "masked_spectral_distance"])



from os.path import join

import seaborn as sns

from ..data.processing.processors import SequenceParsingProcessor
from .postprocessing import normalize_intensity_predictions
from .Report import PDFFile, Report


class IntensityReport(Report):
    """Report generation for Fragment Ion Intensity Prediction tasks."""

    TARGETS_LABEL = "x"
    PREDICTIONS_LABEL = "y"
    DEFAULT_BATCH_SIZE = 600
    PREDICTIONS_COL_NAME = "intensities_pred"

    def __init__(self, output_path, history, figures_ext="png", batch_size=0):
        super(IntensityReport, self).__init__(output_path, history, figures_ext)

        self.pdf_file = PDFFile("DLOmix - Fragment Ion Intensity Report")

        if batch_size:
            self.batch_size = batch_size
        else:
            self.batch_size = IntensityReport.DEFAULT_BATCH_SIZE

    def generate_report(
        self,
        dataset,
        predictions,
        split="test",
        precursor_charge_column_name="precursor_charge_onehot",
    ):
        self._init_report_resources()

        self.plot_all_metrics()

        # make custom plots
        self.plot_spectral_angle(
            dataset, predictions, split, precursor_charge_column_name
        )

        self._compile_report_resources_add_pdf_pages()
        self.pdf_file.output(join(self._output_path, "intensity_Report.pdf"), "F")

    def plot_spectral_angle(
        self, dataset, predictions, split, precursor_charge_column_name
    ):
        """Create spectral  plot

        Arguments
        ---------
        dataset:  FragmentIonIntensityDataset
        predictions:  array of predictions
        split:  str for the split name in the FragmentIonIntensityDataset
        precursor_charge_column_name:  str
        """

        label_column = (
            dataset.label_column
            if isinstance(dataset.label_column, str)
            else dataset.label_column[0]
        )
        model_features = (
            list(dataset.model_features)
            if isinstance(dataset.model_features, list)
            else [dataset.model_features]
        )

        predictions_df = (
            dataset[split]
            .select_columns(
                [
                    SequenceParsingProcessor.PARSED_COL_NAMES["seq"],
                    label_column,
                    *model_features,
                ]
            )
            .to_pandas()
        )

        predictions_df[IntensityReport.PREDICTIONS_COL_NAME] = predictions.tolist()

        predictions_acc = normalize_intensity_predictions(
            predictions_df,
            sequence_column_name=SequenceParsingProcessor.PARSED_COL_NAMES["seq"],
            labels_column_name=label_column,
            predictions_column_name=IntensityReport.PREDICTIONS_COL_NAME,
            precursor_charge_column_name=precursor_charge_column_name,
            batch_size=self.batch_size,
        )

        self.prediction_results = predictions_acc

        violin_plot = sns.violinplot(predictions_acc["spectral_angle"])

        save_path = join(
            self._output_path, "violin_spectral_angle_plot" + self._figures_ext
        )

        fig = violin_plot.get_figure()
        fig.savefig(save_path)

        self._add_report_resource(
            "spectral_angle_plot",
            "Spectral angle plot",
            "The following figure shows the spectral angle plot for the test data.",
            save_path,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..constants import CLASSES_LABELS, padding_char


class DetectabilityModel(nn.Module):
    def __init__(
        self,
        num_units,
        alphabet_size,
        num_classes=len(CLASSES_LABELS),
        padding_idx=padding_char,
    ):
        super(DetectabilityModel, self).__init__()

        self.num_units = num_units
        self.num_classes = num_classes
        self.padding_idx = padding_idx
        self.alphabet_size = alphabet_size

        # Equivalent to Keras one-hot encoding layer
        self.embedding = nn.Embedding(
            alphabet_size, alphabet_size, padding_idx=padding_idx
        )

        self.encoder = Encoder(num_units, alphabet_size)
        self.decoder = Decoder(num_units, num_classes)

    def create_padding_mask(self, x):
        # Create mask where padding_idx tokens are 1 and others are 0
        mask = x == self.padding_idx
        return mask

    def forward(self, x):
        # Create padding mask
        padding_mask = self.create_padding_mask(x)

        # One-hot encoding
        x = self.embedding(x)

        # Encoder
        encoder_outputs, (state_f, state_b) = self.encoder(x, padding_mask)

        # Concatenate forward and backward states
        decoder_hidden = torch.cat([state_f, state_b], dim=-1)

        # Decoder
        output = self.decoder(
            decoder_hidden, state_f, state_b, encoder_outputs, padding_mask
        )

        return output


class Encoder(nn.Module):
    def __init__(self, hidden_size, input_size):
        super(Encoder, self).__init__()

        self.hidden_size = hidden_size

        # Bidirectional GRU
        self.gru = nn.GRU(
            input_size=input_size,
            hidden_size=hidden_size,
            bidirectional=True,
            batch_first=True,
        )

    def forward(self, x, mask):
        # x shape: (batch_size, seq_len, input_size)
        # mask shape: (batch_size, seq_len)

        # Create packed sequence
        lengths = (~mask).sum(dim=1).cpu()  # Get lengths of non-padded sequence
        packed_x = nn.utils.rnn.pack_padded_sequence(
            x, lengths, batch_first=True, enforce_sorted=False
        )

        # Process through GRU
        packed_output, hidden = self.gru(packed_x)

        # Unpack the sequence
        output, _ = nn.utils.rnn.pad_packed_sequence(
            packed_output, batch_first=True, padding_value=0.0
        )

        # Apply mask to output
        mask = mask.unsqueeze(-1).expand(-1, -1, output.size(-1))
        output = output.masked_fill(mask, 0.0)

        return output, hidden


class BahdanauAttention(nn.Module):
    def __init__(self, hidden_size):
        super(BahdanauAttention, self).__init__()

        self.W1 = nn.LazyLinear(hidden_size)
        self.W2 = nn.LazyLinear(hidden_size)
        self.V = nn.LazyLinear(1)

    def forward(self, query, values, mask):
        # query shape: (batch_size, hidden_size)
        # values shape: (batch_size, seq_len, hidden_size)
        # mask shape: (batch_size, seq_len)

        # Add time axis to query
        query = query.unsqueeze(1)

        # Calculate attention scores
        query_values = torch.tanh(self.W1(query) + self.W2(values))
        scores = self.V(query_values)

        # Apply mask to attention scores
        # Set masked positions to -inf before softmax
        mask = mask.unsqueeze(-1)
        scores = scores.masked_fill(mask, float("-inf"))

        # Apply softmax to get attention weights
        attention_weights = F.softmax(scores, dim=1)

        # Apply attention weights to values
        context = attention_weights * values

        # Sum over the time axis
        context = torch.sum(context, dim=1)

        return context


class Decoder(nn.Module):
    def __init__(self, hidden_size, num_classes):
        super(Decoder, self).__init__()

        self.hidden_size = hidden_size
        self.num_classes = num_classes

        self.attention = BahdanauAttention(hidden_size)

        self.gru = nn.GRU(
            input_size=hidden_size * 2,
            hidden_size=hidden_size,
            bidirectional=True,
            batch_first=True,
        )

        self.dense = nn.Linear(hidden_size * 2, num_classes)

    def forward(self, decoder_input, state_f, state_b, encoder_outputs, mask):
        # Apply attention
        context = self.attention(decoder_input, encoder_outputs, mask)

        # Add sequence dimension
        context = context.unsqueeze(1)

        # Pass through GRU
        states = torch.stack([state_f, state_b])
        output, hidden = self.gru(context, states)

        # Pass through final dense layer
        output = self.dense(output)
        output = F.softmax(output, dim=-1)

        return output.squeeze(1)



import tensorflow as tf
import tensorflow.keras.backend as K

# Parts of the code adopted and modified based on:
# https://github.com/horsepurve/DeepRTplus/blob/cde829ef4bd8b38a216d668cf79757c07133b34b/RTdata_emb.py


@tf.keras.utils.register_keras_serializable(package="dlomix")
class TimeDeltaMetric(tf.keras.metrics.Metric):
    """
    Implementation of the time delta metric as a Keras Metric using subclassing.

    Parameters
    ----------
    mean : int, optional
        Mean value of the targets in case normalization was performed. Defaults to 0.
    std : int, optional
        Standard deviation value of the targets in case normalization was performed. Defaults to 1.
    percentage : float, optional
        What percentage of the data points to consider, this is specific to the computation of the metric. Defaults to 0.95 which corresponds to 95% of the data points and is the mostly used value in papers.
    name : str, optional
        Name of the metric so that it can be reported and used later in Keras History objects. Defaults to 'timedelta'.
    rescale_targets : bool, optional
        Whether to rescale (denormalize) targets or not. Defaults to False.
    rescale_predictions : bool, optional
        Whether to rescale (denormalize) predictions or not. Defaults to False.
    double_delta : bool, optional
        Whether to multiply the computed delta by 2 in order to make it two-sided or not. Defaults to False.
    """

    def __init__(self, percentage=0.95, name="timedelta", double_delta=False, **kwargs):
        super(TimeDeltaMetric, self).__init__(name=name, **kwargs)
        self.delta = self.add_weight(name="delta", initializer="zeros")
        self.batch_count = self.add_weight(name="batch-count", initializer="zeros")
        self.percentage = percentage
        self.double_delta = double_delta

    def update_state(
        self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight=None
    ) -> None:
        """
        Update the metric state.

        Parameters
        ----------
        y_true : tf.Tensor
            True values of the target.
        y_pred : tf.Tensor
            Predicted values of the target.
        sample_weight : tf.Tensor, optional
            Sample weights. Defaults to None.
        """

        # Compute absolute errors
        abs_error = tf.abs(y_true - y_pred)

        # Sort the errors in ascending order
        sorted_error = tf.sort(abs_error)

        # Find the index corresponding to the 95th percentile
        # First, cast the shape to float, then multiply by 0.95, and finally cast back to int32 for indexing
        percentile_index = tf.cast(
            tf.cast(tf.shape(sorted_error)[0], tf.float32) * self.percentage,
            dtype=tf.int32,
        )

        # Select the 95th percentile value
        delta_value = sorted_error[
            percentile_index - 1
        ]  # tf.shape gives a 0-based index

        # two-sided delta
        if self.double_delta:
            delta_value = delta_value * 2

        # Update the count of batches
        self.batch_count.assign_add(1.0)

        # Update delta (to track sum of deltas over batches)
        self.delta.assign_add(delta_value)

    def result(self):
        # Return the average of deltas over batches
        return self.delta / self.batch_count

    def reset_states(self):
        # Reset the state at the beginning of each epoch
        self.delta.assign(0.0)
        self.batch_count.assign(0.0)

    def get_config(self):
        return {
            "percentage": self.percentage,
            "double_delta": self.double_delta,
            "name": self.name,
        }


@tf.keras.utils.register_keras_serializable("dlomix")
def timedelta(y_true, y_pred, normalize=False, percentage=0.95):
    """
    A functional implementation to compute the 95th percentile of the absolute error between true and predicted values.
    Parameters
    ----------
    y_true : tf.Tensor
        True values of the target.
    y_pred : tf.Tensor
        Predicted values of the target.
    normalize : bool, optional
        Whether to normalize the delta by the range of the true values. Defaults to False.
    percentage : float, optional
        What percentage of the data points to consider, this is specific to the computation of the metric. Defaults to 0.95 which corresponds to 95% of the data points and is the mostly used value in papers.
    Returns
    -------
    tf.Tensor
        The 95th percentile of the absolute error.
    """

    mark_percentile = tf.cast(
        tf.cast(tf.shape(y_true)[0], dtype=tf.float32) * percentage, dtype=tf.int32
    )

    abs_error = K.abs(y_true - y_pred)
    delta = tf.sort(abs_error)[mark_percentile - 1]

    if normalize:
        norm_range = K.max(y_true) - K.min(y_true)
        return (delta) / (norm_range)
    return delta



# -*- coding: utf-8 -*-

import os
from itertools import cycle
from os.path import join

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.metrics import ConfusionMatrixDisplay, auc, confusion_matrix, roc_curve

from ..constants import CLASSES_LABELS
from .Report import PDFFile, Report


class DetectabilityReport(Report):
    """Report generation for Detectability Prediction tasks."""

    def __init__(
        self,
        targets,
        predictions,
        input_data_df,
        output_path,
        history,
        rank_by_prot=False,
        threshold=None,
        figures_ext="png",
        name_of_dataset="unspecified",
        name_of_model="unspecified",
    ):
        super(DetectabilityReport, self).__init__(output_path, history, figures_ext)

        self.pdf_file = PDFFile("DLOmix - Detectability Report")

        self.predictions = predictions
        self.test_size = self.predictions.shape[
            0
        ]  # Removing the last part of the test data which don't fit the batch size
        self.targets = targets[: self.test_size]
        self.input_data_df = input_data_df.loc[: self.test_size - 1]
        self.output_path = output_path
        self.rank_by_prot = rank_by_prot
        self.threshold = threshold
        self.name_of_dataset = name_of_dataset
        self.name_of_model = name_of_model
        self.results_metrics_dict = None
        self.results_report_df = None
        self.detectability_report_table = None

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)

        self.results_dict_and_df()

    def generate_report(self, **kwargs):
        self._init_report_resources()

        self._add_report_resource(
            "name of dataset",
            "Dataset",
            f"The dataset used is {self.name_of_dataset}\n",
            self.name_of_dataset,
        )

        self._add_report_resource(
            "name of model",
            "Model",
            f"The model used to make the prediction is {self.name_of_model}\n",
            self.name_of_model,
        )
        self._add_report_resource(
            "binary_accuracy",
            "Binary accuracy",
            f"The Binary Accuracy value for the predictions is {round(self.results_metrics_dict['binary_accuracy'], 4)}\n",
            self.results_metrics_dict["binary_accuracy"],
        )

        self._add_report_resource(
            "categorical_accuracy",
            "Categorical Accuracy",
            f"The Categorical Accuracy value for the predictions is {round(self.results_metrics_dict['categorical_accuracy'], 4)}\n",
            self.results_metrics_dict["categorical_accuracy"],
        )

        self._add_report_resource(
            "true_positive_rate",
            "True Positive Rate (Recall)",
            f"The True Positive Rate (Recall) value for the predictions is {round(self.results_metrics_dict['true_positive_rate'], 4)}\n",
            self.results_metrics_dict["true_positive_rate"],
        )

        self._add_report_resource(
            "false_positive_rate",
            "False Positive Rate (Specificity)",
            f"The False Positive Rate (Specificity) value for the predictions is {round(self.results_metrics_dict['false_positive_rate'], 4)}\n",
            self.results_metrics_dict["false_positive_rate"],
        )

        self._add_report_resource(
            "precision",
            "Precision",
            f"The Presicion value for the predictions is {round(self.results_metrics_dict['precision'], 4)}\n",
            self.results_metrics_dict["precision"],
        )

        self._add_report_resource(
            "f1_score",
            "F1 Score",
            f"The F1 Score value for the predictions is {round(self.results_metrics_dict['f1_score'], 4)}\n",
            self.results_metrics_dict["f1_score"],
        )

        self._add_report_resource(
            "MCC",
            "Matthews Correlation Coefficient (MCC)",
            f"The Matthews Correlation Coefficient (MCC) value for the predictions is {round(self.results_metrics_dict['MCC'], 4)}\n",
            self.results_metrics_dict["MCC"],
        )

        self.plot_all_metrics()
        self.plot_roc_curve_binary()
        self.plot_confusion_matrix_binary()
        self.plot_roc_curve()
        self.plot_confusion_matrix_multiclass()
        self.plot_heatmap_prediction_prob_error()
        self._compile_report_resources_add_pdf_pages()
        self.pdf_file.output(
            join(self._output_path, "Detectability_evaluation_report.pdf"), "F"
        )

    def results_dict_and_df(self):
        eval_result = evaluation_results(
            self.predictions, self.targets, threshold=self.threshold
        )
        self.results_metrics_dict = eval_result.eval_results

        target_labels = {
            0: "Non-Flyer",
            1: "Weak Flyer",
            2: "Intermediate Flyer",
            3: "Strong Flyer",
        }
        binary_labels = {0: "Non-Flyer", 1: "Flyer"}

        df_data_results = self.input_data_df.copy().reset_index(drop=True)
        df_data_results = pd.concat(
            [
                df_data_results,
                pd.DataFrame(
                    self.results_metrics_dict["predictions"], columns=["Predictions"]
                ),
            ],
            axis=1,
        )

        for i, label in enumerate(CLASSES_LABELS):
            df_data_results[label] = np.round_(
                np.array(self.results_metrics_dict["probabilities"])[:, i], decimals=3
            )

        df_data_results["Flyer"] = (
            df_data_results["Weak Flyer"]
            + df_data_results["Intermediate Flyer"]
            + df_data_results["Strong Flyer"]
        )
        # df_data_results['Flyer'] = round(df_data_results['Flyer'], ndigits = 3)
        df_data_results["Binary Predictions"] = np.where(
            df_data_results["Predictions"] == 0, 0, 1
        )
        df_data_results["Binary Classes"] = np.where(
            df_data_results["Classes"] == 0, 0, 1
        )

        sorted_columns = [
            "Sequences",
            "Proteins",
            "Weak Flyer",
            "Intermediate Flyer",
            "Strong Flyer",
            "Non-Flyer",
            "Flyer",
            "Classes",
            "Predictions",
            "Binary Classes",
            "Binary Predictions",
        ]

        all_columns = [x for x in sorted_columns if x in df_data_results.columns]

        df_data_results = df_data_results[all_columns]
        df_final_results = df_data_results.copy()

        if "Proteins" in df_final_results.columns and self.rank_by_prot:
            df_final_results = df_final_results.sort_values(
                by=[
                    "Proteins",
                    "Flyer",
                    "Predictions",
                    "Strong Flyer",
                    "Intermediate Flyer",
                    "Weak Flyer",
                ],
                ascending=[True, False, False, False, False, False],
            ).reset_index(drop=True)

            df_final_results["Rank"] = (
                df_final_results.groupby("Proteins")["Flyer"]
                .rank(ascending=False, method="first")
                .astype(int)
            )
        #           df_final_results['Rank_2'] = df_final_results.groupby('Proteins')['Flyer'].rank(ascending = False, method = 'dense').astype(int)

        else:
            df_final_results = df_final_results.sort_values(
                by=[
                    "Flyer",
                    "Predictions",
                    "Strong Flyer",
                    "Intermediate Flyer",
                    "Weak Flyer",
                ],
                ascending=[False, False, False, False, False],
            ).reset_index(drop=True)

            df_final_results["Rank"] = (
                df_final_results["Flyer"]
                .rank(ascending=False, method="first")
                .astype(int)
            )
        #            df_final_results['Rank_2'] = df_final_results['Flyer'].rank(ascending = False, method = 'dense').astype(int)

        df_final_results["Classes"] = df_final_results["Classes"].map(target_labels)
        df_final_results["Binary Classes"] = df_final_results["Binary Classes"].map(
            binary_labels
        )
        df_final_results["Predictions"] = df_final_results["Predictions"].map(
            target_labels
        )
        df_final_results["Binary Predictions"] = df_final_results[
            "Binary Predictions"
        ].map(binary_labels)

        self.results_report_df = df_data_results
        self.detectability_report_table = df_final_results
        save_path_ = join(self.output_path, "Dectetability_prediction_report.csv")
        self.detectability_report_table.to_csv(save_path_, index=False)

    def plot_roc_curve_binary(self):
        """Plot ROC curve (Binary classification)

        Arguments
        ----------
            binary_targets: Array with binary target values
            binary_predictions_prob: Array with binary prediction probability values
        """

        fpr, tpr, thresholds = roc_curve(
            np.array(self.results_report_df["Binary Classes"]),
            np.array(self.results_report_df["Flyer"]),
        )
        AUC_score = auc(fpr, tpr)

        # create ROC curve

        plt.plot(fpr, tpr)
        plt.title(
            "Receiver operating characteristic curve (Binary classification)",
            y=1.04,
            fontsize=10,
        )
        plt.ylabel("True Positive Rate")
        plt.xlabel("False Positive Rate")
        save_path = join(
            self._output_path, "ROC_curve_binary_classification" + self._figures_ext
        )

        plt.savefig(save_path, bbox_inches="tight", dpi=90)
        plt.show()
        plt.close()

        self._add_report_resource(
            "roc_curve_binary",
            "ROC curve (Binary classification)",
            "The following plot shows the Receiver operating characteristic (ROC) curve for the binary classification.",
            save_path,
        )

        self._add_report_resource(
            "AUC_binary_score",
            "AUC Binary Score",
            f"The AUC score value for the binary classification is {round(AUC_score, 4)}",
            AUC_score,
        )

    def plot_confusion_matrix_binary(self):
        """Plot confusion matrix (Binary classification)

        Arguments
        ----------
            binary_targets: Array-like of shape (n_samples,) with binary target values
            binary_predictions_prob: Array-like of shape (n_samples,) with binary prediction classes (not probabilities) values
        """
        conf_matrix = confusion_matrix(
            self.results_report_df["Binary Classes"],
            self.results_report_df["Binary Predictions"],
        )

        conf_matrix
        conf_matrix_disp = ConfusionMatrixDisplay(
            confusion_matrix=conf_matrix, display_labels=["Non-Flyer", "Flyer"]
        )
        fig, ax = plt.subplots()
        conf_matrix_disp.plot(xticks_rotation=45, ax=ax)
        plt.title("Confusion Matrix (Binary Classification)", y=1.04, fontsize=11)
        save_path = join(
            self._output_path, "confusion_matrix_binary" + self._figures_ext
        )
        plt.savefig(save_path, bbox_inches="tight", dpi=80)
        plt.show()
        plt.close()

        self._add_report_resource(
            "confusion_matrix_binary",
            "Confusion Matrix (Binary Classification)",
            "The following plot shows the Confusion Matrix (Binary Classification).",
            save_path,
        )

    def plot_roc_curve(self):
        """Plot ROC curve (Multiclass classification)

        Arguments
        ----------
            multiclass_targets: Array with multiclass targets values
            multiclass_predictions_prob: Array with multiclass prediction probability values
        """
        # Compute ROC curve and ROC area for each class
        fpr = {}
        tpr = {}
        roc_auc = {}

        for i in range(len(CLASSES_LABELS)):
            fpr[i], tpr[i], _ = roc_curve(
                np.squeeze(self.targets)[:, i], np.squeeze(self.predictions)[:, i]
            )
            roc_auc[i] = auc(fpr[i], tpr[i])

        lw = 2
        colors = cycle(["blue", "red", "green", "gold"])
        for i, color in zip(range(len(CLASSES_LABELS)), colors):
            plt.plot(
                fpr[i],
                tpr[i],
                color=color,
                lw=2,
                label="ROC curve of class {0} (area = {1:0.2f})"
                "".format(CLASSES_LABELS[i], roc_auc[i]),
            )

        plt.plot([0, 1], [0, 1], "k--", lw=lw)
        plt.xlim([-0.05, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel("False Positive Rate")
        plt.ylabel("True Positive Rate")
        plt.title("ROC curve for multi-class data", y=1.04, fontsize=10)
        plt.legend(loc="best", fontsize="small")
        save_path = join(
            self._output_path, "ROC_curve_multiclass_classification" + self._figures_ext
        )
        plt.savefig(save_path, bbox_inches="tight", dpi=90)
        plt.show()
        plt.close()

        self._add_report_resource(
            "roc_curve_multiclass",
            "ROC curve (Multiclass classification)",
            "The following plot shows the Receiver operating characteristic (ROC) curve for the multiclass classification.",
            save_path,
        )

        self._add_report_resource(
            "AUC_multiclass_score",
            "AUC Multiclass Score",
            f"The AUC score value for the multiclass classification is: {CLASSES_LABELS[0]}: {round(roc_auc[0], 4)}, {CLASSES_LABELS[1]}: {round(roc_auc[1], 4)},\
            {CLASSES_LABELS[2]}: {round(roc_auc[2], 4)}, {CLASSES_LABELS[3]}: {round(roc_auc[3], 4)}.",
            roc_auc,
        )

    def plot_confusion_matrix_multiclass(self):
        """Plot confusion matrix (Multiclass classification)

        Arguments
        ----------
            multiclass_targets: Array-like of shape (n_samples,) with multiclass target values
            multiclass_predictions: Array-like of shape (n_samples,) with multiclass prediction classes (not probabilities) values
        """

        multi_conf_matrix = confusion_matrix(
            self.results_report_df["Classes"], self.results_report_df["Predictions"]
        )

        conf_matrix_disp = ConfusionMatrixDisplay(
            confusion_matrix=multi_conf_matrix, display_labels=CLASSES_LABELS
        )  #
        fig, ax = plt.subplots()
        conf_matrix_disp.plot(xticks_rotation=45, ax=ax)
        plt.title(
            "Confusion Matrix (Multiclass Classification)", y=1.04, fontsize=11
        )  # , y=1.12
        save_path = join(
            self._output_path, "confusion_matrix_multiclass" + self._figures_ext
        )
        plt.savefig(save_path, bbox_inches="tight", dpi=80)
        plt.show()
        plt.close()

        self._add_report_resource(
            "confusion_matrix_multiclass",
            "Confusion Matrix (Multiclass Classification)",
            "The following plot shows the Confusion Matrix (Multiclass Classification).",
            save_path,
        )

    def plot_heatmap_prediction_prob_error(self):
        """Plot Heatmap of average error between probabilities of real classes vs predicted

        Arguments
        ----------
            dict_of_prob_difference: Dictionary containing the average difference between the predicted probabilities of
            the predicted classes and the predicted probabilities of real classes

        """
        probability_var = {}
        # probability_var_with_std = {}

        for k, v in self.results_metrics_dict["delta_prob_pred"].items():
            probability_var[k] = {}
            # probability_var_with_std[k] = {}

            for m, n in self.results_metrics_dict["delta_prob_pred"][k].items():
                # probability_var_with_std[k][m] = {}

                probability_var[k][m] = round(
                    np.mean(self.results_metrics_dict["delta_prob_pred"][k][m]),
                    ndigits=3,
                )

        #                 probability_var_with_std[k][m]['mean'] = round(np.mean(self.results_metrics_dict['delta_prob_pred'][k][m]), ndigits = 3)
        #                 probability_var_with_std[k][m]['std'] = round(np.std(self.results_metrics_dict['delta_prob_pred'][k][m]), ndigits = 3)

        df_probability_var = pd.DataFrame(probability_var)  # \
        df_probability_var.columns = CLASSES_LABELS
        df_probability_var.index = CLASSES_LABELS
        sns.heatmap(df_probability_var, cmap="viridis", linewidths=0.05, annot=True)
        plt.yticks(rotation=0)
        plt.xticks(rotation=45)
        plt.title(
            "Heatmap of average error between probabilities of real classes vs predicted",
            y=1.04,
            fontsize=11,
        )  # , y=1.12
        save_path = join(
            self._output_path, "heatmap_prediction_prob_error" + self._figures_ext
        )
        plt.savefig(save_path, bbox_inches="tight", dpi=80)
        plt.show()
        plt.close()

        self._add_report_resource(
            "heatmap_prediction_prob_error",
            "Average error between probabilities of real classes vs predicted",
            "The following plot shows the average error between probabilities of real classes vs predicted.",
            save_path,
        )


class evaluation_results:
    def __init__(
        self,
        predictions,
        targets,
        num_clases=len(CLASSES_LABELS),
        threshold=None,
        print_results=True,
    ):
        super(evaluation_results, self).__init__()

        self.predictions = predictions
        self.targets = targets
        self.num_clases = num_clases
        self.threshold = threshold
        self.print_results = print_results
        self.all_pred = None

        self.correct_p = 0
        self.incorrect_p = 0

        self.TP = 0
        self.TN = 0
        self.FP = 0
        self.FN = 0

        self.results_dict = {}
        self.diff_prob_dict = {}
        self.eval_results = {}
        self.evaluate()

    def evaluate(self):
        predicted = np.empty(len(self.predictions))

        for i in range(self.num_clases):
            I = str(i)

            self.results_dict[I] = {}
            self.diff_prob_dict[I] = {}

            for j in range(self.num_clases):
                J = str(j)

                self.results_dict[I][J] = 0
                self.diff_prob_dict[I][J] = []

        if self.threshold:
            thresh_pred = np.squeeze(self.predictions)

            index_1 = thresh_pred[:, 0] >= self.threshold
            index_2 = thresh_pred[:, 0] < self.threshold

            predicted[index_1] = 0
            predicted[index_2] = np.argmax(thresh_pred[:, 1:][index_2], axis=-1) + 1

        else:
            predicted = np.argmax(self.predictions, axis=-1)

        probabilities = np.squeeze(self.predictions)

        expected = self.targets

        expected_str = np.squeeze([str(x) for x in np.argmax(expected, axis=-1)])
        expected_int = np.squeeze(np.argmax(expected, axis=-1))
        predicted_str = np.array([str(int(x)) for x in predicted])
        predicted = np.array([int(x) for x in predicted])

        self.all_pred = predicted

        for i in range(len(predicted)):
            self.results_dict[expected_str[i]][predicted_str[i]] += 1

            diff_prob = np.absolute(
                probabilities[i, expected_int[i]] - probabilities[i, predicted[i]]
            )

            self.diff_prob_dict[expected_str[i]][predicted_str[i]].append(diff_prob)

        correct_p = sum(expected_int == predicted)

        incorrect_p = sum(expected_int != predicted)

        non_flyer_index = np.array(expected_int == 0)

        self.TN = int(sum(expected_int[non_flyer_index] == predicted[non_flyer_index]))

        self.FP = int(sum(expected_int[non_flyer_index] != predicted[non_flyer_index]))

        flyer_index = np.array(expected_int != 0)

        self.TP = int(sum(predicted[flyer_index] != 0))

        self.FN = int(sum(predicted[flyer_index] == 0))

        if (self.TP + self.TN + self.FP + self.FN) > 0:
            binary_accuracy = (self.TP + self.TN) / (
                self.TP + self.TN + self.FP + self.FN
            )
        else:
            binary_accuracy = None

        if (correct_p + incorrect_p) > 0:
            overall_accuracy = correct_p / (correct_p + incorrect_p)
        else:
            overall_accuracy = None

        if (self.TP + self.FN) > 0:
            true_positive_rate = self.TP / (self.TP + self.FN)
        else:
            true_positive_rate = None

        if (self.TN + self.FP) > 0:
            false_positive_rate = self.TN / (self.TN + self.FP)
        else:
            false_positive_rate = None

        if (self.TP + self.FP) > 0:
            precision = self.TP / (self.TP + self.FP)
        else:
            precision = None

        if precision > 0 and true_positive_rate > 0:
            f_score = 2 / ((1 / precision) + (1 / true_positive_rate))
        else:
            f_score = None

        if (
            (self.TP + self.FP)
            * (self.TP + self.FN)
            * (self.TN + self.FP)
            * (self.TN + self.FN)
        ) > 0:
            MCC = ((self.TP * self.TN) - (self.FP * self.FN)) / np.sqrt(
                float(
                    (self.TP + self.FP)
                    * (self.TP + self.FN)
                    * (self.TN + self.FP)
                    * (self.TN + self.FN)
                )
            )
        else:
            MCC = None

        conf_matrix = {"TP": self.TP, "TN": self.TN, "FP": self.FP, "FN": self.FN}

        self.eval_results = {
            "predictions": predicted,
            "probabilities": probabilities,
            "categorical_accuracy": overall_accuracy,
            "binary_accuracy": binary_accuracy,
            "true_positive_rate": true_positive_rate,
            "false_positive_rate": false_positive_rate,
            "precision": precision,
            "f1_score": f_score,
            "MCC": MCC,
            "conf_matrix": conf_matrix,
            "results_dict": self.results_dict,
            "delta_prob_pred": self.diff_prob_dict,
        }

        if self.print_results:
            print(
                f'Binary Accuracy: {round(self.eval_results["binary_accuracy"], ndigits = 2)}'
            )

            print(
                f'\nCategorical Accuracy: {round(self.eval_results["categorical_accuracy"], ndigits = 2)}'
            )

            print(
                f'\nMatthews Correlation Coefficient (MCC): {round(self.eval_results["MCC"], ndigits = 2)}'
            )

            print(
                f'\nTrue Positive Rate (Recall): {round(self.eval_results["true_positive_rate"], ndigits = 2)}'
            )

            print(
                f'\nFalse Positive Rate (Specificity): {round(self.eval_results["false_positive_rate"], ndigits = 2)}'
            )

            print(f'\nPrecision: {round(self.eval_results["precision"], ndigits = 2)}')

            print(f'\nF1 Score: {round(self.eval_results["f1_score"], ndigits = 2)}')


class predictions_report:
    def __init__(
        self,
        predictions,
        input_data_df,
        output_path,
        num_clases=len(CLASSES_LABELS),
        rank_by_prot=False,
        threshold=None,
    ):
        super(predictions_report, self).__init__()

        self.predictions = np.squeeze(predictions)
        self.test_size = self.predictions.shape[
            0
        ]  # Removing the last part of the test data which don't fit the batch size
        self.input_data_df = input_data_df.loc[: self.test_size - 1]
        self.output_path = output_path
        self.num_clases = num_clases
        self.rank_by_prot = rank_by_prot
        self.threshold = threshold

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)

        self.all_pred = None
        self.predictions_report = None
        self.evaluate()

    def evaluate(self):
        predicted = np.empty(len(self.predictions))

        target_labels = {
            0: "Non-Flyer",
            1: "Weak Flyer",
            2: "Intermediate Flyer",
            3: "Strong Flyer",
        }
        binary_labels = {0: "Non-Flyer", 1: "Flyer"}

        if self.threshold:
            thresh_pred = np.squeeze(self.predictions)

            index_1 = thresh_pred[:, 0] >= self.threshold
            index_2 = thresh_pred[:, 0] < self.threshold

            predicted[index_1] = 0
            predicted[index_2] = np.argmax(thresh_pred[:, 1:][index_2], axis=-1) + 1

        else:
            predicted = np.argmax(self.predictions, axis=-1)
            predicted = [x for x in predicted]

        self.all_pred = predicted

        df_data_results = self.input_data_df.copy().reset_index(drop=True)
        df_data_results = pd.concat(
            [df_data_results, pd.DataFrame(self.all_pred, columns=["Predictions"])],
            axis=1,
        )

        for i, label in enumerate(CLASSES_LABELS):
            df_data_results[label] = np.round_(
                np.array(self.predictions)[:, i], decimals=3
            )

        df_data_results["Flyer"] = (
            df_data_results["Weak Flyer"]
            + df_data_results["Intermediate Flyer"]
            + df_data_results["Strong Flyer"]
        )
        # df_data_results['Flyer'] = round(df_data_results['Flyer'], ndigits = 3)
        df_data_results["Flyer"] = np.round_(df_data_results["Flyer"], decimals=3)
        df_data_results["Binary Predictions"] = np.where(
            df_data_results["Predictions"] == 0, 0, 1
        )

        if "Classes" in df_data_results.columns:
            df_data_results["Binary Classes"] = np.where(
                df_data_results["Classes"] == 0, 0, 1
            )
            df_data_results["Classes"] = df_data_results["Classes"].map(target_labels)
            df_data_results["Binary Classes"] = df_data_results["Binary Classes"].map(
                binary_labels
            )

        sorted_columns = [
            "Sequences",
            "Proteins",
            "Weak Flyer",
            "Intermediate Flyer",
            "Strong Flyer",
            "Non-Flyer",
            "Flyer",
            "Classes",
            "Predictions",
            "Binary Classes",
            "Binary Predictions",
        ]

        all_columns = [x for x in sorted_columns if x in df_data_results.columns]

        df_data_results = df_data_results[all_columns]

        if "Proteins" in df_data_results.columns and self.rank_by_prot:
            df_data_results = df_data_results.sort_values(
                by=[
                    "Proteins",
                    "Flyer",
                    "Predictions",
                    "Strong Flyer",
                    "Intermediate Flyer",
                    "Weak Flyer",
                ],
                ascending=[True, False, False, False, False, False],
            ).reset_index(drop=True)

            df_data_results["Rank"] = (
                df_data_results.groupby("Proteins")["Flyer"]
                .rank(ascending=False, method="first")
                .astype(int)
            )
        #            df_data_results['Rank_2'] = df_data_results.groupby('Proteins')['Flyer'].rank(ascending = False, method = 'dense').astype(int)

        else:
            df_data_results = df_data_results.sort_values(
                by=[
                    "Flyer",
                    "Predictions",
                    "Strong Flyer",
                    "Intermediate Flyer",
                    "Weak Flyer",
                ],
                ascending=[False, False, False, False, False],
            ).reset_index(drop=True)

            df_data_results["Rank"] = (
                df_data_results["Flyer"]
                .rank(ascending=False, method="first")
                .astype(int)
            )
        #            df_data_results['Rank_2'] = df_data_results['Flyer'].rank(ascending = False, method = 'dense').astype(int)

        df_data_results["Predictions"] = df_data_results["Predictions"].map(
            target_labels
        )
        df_data_results["Binary Predictions"] = df_data_results[
            "Binary Predictions"
        ].map(binary_labels)

        self.predictions_report = df_data_results

        save_path = join(self.output_path, "Dectetability_prediction_report.csv")
        self.predictions_report.to_csv(save_path, index=False)



from ..config import _BACKEND, PYTORCH_BACKEND, TENSORFLOW_BACKEND

__all__ = []

if _BACKEND in TENSORFLOW_BACKEND:
    from .cyclic_lr import CyclicLR

    __all__.append("CyclicLR")

elif _BACKEND in PYTORCH_BACKEND:
    pass



import logging

import numpy as np

from dlomix.data.processing.feature_extractors import LookupFeatureExtractor

logger = logging.getLogger(__name__)


def test_lookup_feature_extractor_exact_length(lookup_table):
    # Create a sequence of indices to look up
    sequence_for_lookup = [0, 1, 3]
    sequence_length = len(sequence_for_lookup)
    max_length = 3
    default_value = [-1.0, -1.0]

    # Create the feature extractor
    feature_extractor = LookupFeatureExtractor(
        sequence_column_name="sequence",
        feature_column_name="feature",
        lookup_table=lookup_table,
        feature_default_value=default_value,
        max_length=max_length,
    )

    # Extract features for the given sequence
    feature = feature_extractor._extract_feature(sequence_for_lookup)
    logger.info("Extracted feature:\n%s", feature)

    assert feature.shape == (
        max_length,
        2,
    ), f"Expected feature shape to be ({max_length}, 2), but got {feature.shape}"

    assert np.array_equal(
        feature[:sequence_length],
        np.array([lookup_table[idx] for idx in sequence_for_lookup]),
    ), "The extracted feature values do not match the expected values from the lookup table."

    logger.info(feature[sequence_length:])
    assert (
        len(feature[sequence_length:]) == 0
    ), "The padded feature values do not match the expected default value from the lookup table."


def test_lookup_feature_extractor_with_padding(lookup_table):
    # Create a sequence of indices to look up
    sequence_for_lookup = [0, 1, 3]
    sequence_length = len(sequence_for_lookup)
    max_length = 6
    default_value = [-5.0, -5.0]

    # Create the feature extractor
    feature_extractor = LookupFeatureExtractor(
        sequence_column_name="sequence",
        feature_column_name="feature",
        lookup_table=lookup_table,
        feature_default_value=default_value,
        max_length=max_length,
    )

    # Extract features for the given sequence
    feature = feature_extractor._extract_feature(sequence_for_lookup)
    logger.info("Extracted feature:\n%s", feature)

    assert feature.shape == (
        max_length,
        2,
    ), f"Expected feature shape to be ({max_length}, 2), but got {feature.shape}"

    assert np.array_equal(
        feature[:sequence_length],
        np.array([lookup_table[idx] for idx in sequence_for_lookup]),
    ), "The extracted feature values do not match the expected values from the lookup table."

    assert np.array_equal(
        feature[sequence_length:],
        np.array([default_value] * (max_length - sequence_length)),
    ), "The padded feature values do not match the expected default value from the lookup table."



import logging

import torch

from dlomix.layers.attention_torch import AttentionLayer, DecoderAttentionLayer

logger = logging.getLogger(__name__)


def test_attention_layer_shapes():
    batch_size = 2
    seq_len = 5
    feature_dim = 8

    dummy_input = torch.randn(batch_size, seq_len, feature_dim)

    layer = AttentionLayer(
        feature_dim=feature_dim, seq_len=seq_len, context=False, bias=True
    )
    output = layer(dummy_input)

    logger.info(
        "Original Shape: {}, Layer output shape".format(
            (batch_size, seq_len, feature_dim), output.shape
        )
    )

    assert output.detach().numpy().shape == (batch_size, feature_dim)


def test_decoder_attention_layer_shapes():
    batch_size = 2
    time_steps = 30
    features = 5

    dummy_input = torch.randn(batch_size, time_steps, features)
    layer = DecoderAttentionLayer(time_steps=time_steps)
    output = layer(dummy_input)

    logger.info(
        "Original Shape: {}, Layer output shape".format(
            (batch_size, time_steps, features), output.shape
        )
    )

    assert output.detach().numpy().shape == (batch_size, time_steps, features)



import logging

import tensorflow as tf

from dlomix.losses.intensity import (
    masked_pearson_correlation_distance,
    masked_spectral_distance,
)

logger = logging.getLogger(__name__)


# ------------------ intensity - masked spectral distance ------------------


def test_spectral_distance_identical():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for identical vectors: {}".format(sa.numpy()))

    assert sa.numpy() == 0


def test_spectral_distance_different():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor([list(reversed(y_true[0]))])

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for reversed vectors: {}".format(sa.numpy()))

    assert sa.numpy() != 0


def test_spectral_distance_zero_input():
    y_true = [[0.0, 0.0, 0.0]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for zero input vectors: {}".format(sa.numpy()))


# ------------------ intensity - masked pearson correlation distance ------------------


def test_pearson_correlation_distance_identical():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    pc = masked_pearson_correlation_distance(y_true_tensor, y_pred_tensor)
    logger.info(
        "Masked Pearson Correlation Distance for identical vectors: {}".format(
            pc.numpy()
        )
    )

    assert pc.numpy() == 0


def test_pearson_correlation_distance_different():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor([list(reversed(y_true[0]))])

    pc = masked_pearson_correlation_distance(y_true_tensor, y_pred_tensor)
    logger.info(
        "Masked Pearson Correlation Distance for reversed vectors: {}".format(
            pc.numpy()
        )
    )

    assert pc.numpy() != 0


def test_pearson_correlation_distance_zero_input():
    y_true = [[0.0, 0.0, 0.0]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    pc = masked_pearson_correlation_distance(y_true_tensor, y_pred_tensor)
    logger.info(
        "Masked Pearson Correlation Distance for zero input vectors: {}".format(
            pc.numpy()
        )
    )


# ---------------------------------------------------------------------------



import numpy as np
import tensorflow as tf


@tf.keras.utils.register_keras_serializable("dlomix")
class CyclicLR(tf.keras.callbacks.Callback):
    """This callback implements a cyclical learning rate policy (CLR).
    The method cycles the learning rate between two boundaries with
    some constant frequency, as detailed in this paper (https://arxiv.org/abs/1506.01186).

    """

    def __init__(
        self,
        base_lr=0.001,
        max_lr=0.006,
        step_size=2000.0,
        mode="triangular",
        gamma=1.0,
        scale_fn=None,
        scale_mode="cycle",
    ):
        super(CyclicLR, self).__init__()

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.mode = mode
        self.gamma = gamma

        if scale_fn is None:
            if self.mode == "triangular":
                self.scale_fn = lambda x: 1.0
                self.scale_mode = "cycle"
            elif self.mode == "triangular2":
                self.scale_fn = lambda x: 1 / (2.0 ** (x - 1))
                self.scale_mode = "cycle"
            elif self.mode == "exp_range":
                self.scale_fn = lambda x: gamma ** (x)
                self.scale_mode = "iterations"
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode

        self.clr_iterations = 0.0
        self.trn_iterations = 0.0
        self.history = {}

        self._reset()

    def _reset(self, new_base_lr=None, new_max_lr=None, new_step_size=None):
        """Resets cycle iterations.
        Optional boundary/step size adjustment.
        """
        if new_base_lr is not None:
            self.base_lr = new_base_lr
        if new_max_lr is not None:
            self.max_lr = new_max_lr
        if new_step_size is not None:
            self.step_size = new_step_size
        self.clr_iterations = 0.0

    def clr(self):
        cycle = np.floor(1 + self.clr_iterations / (2 * self.step_size))
        x = np.abs(self.clr_iterations / self.step_size - 2 * cycle + 1)
        if self.scale_mode == "cycle":
            return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(
                0, (1 - x)
            ) * self.scale_fn(cycle)
        else:
            return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(
                0, (1 - x)
            ) * self.scale_fn(self.clr_iterations)

    def on_train_begin(self, logs=None):

        if self.clr_iterations == 0:
            tf.keras.backend.set_value(self.model.optimizer.lr, self.base_lr)
        else:
            tf.keras.backend.set_value(self.model.optimizer.lr, self.clr())

    def on_batch_end(self, epoch, logs=None):
        logs = logs or {}
        self.trn_iterations += 1
        self.clr_iterations += 1

        self.history.setdefault("lr", []).append(
            tf.keras.backend.get_value(self.model.optimizer.lr)
        )
        self.history.setdefault("iterations", []).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        tf.keras.backend.set_value(self.model.optimizer.lr, self.clr())

    def get_config(self):
        return {
            "base_lr": self.base_lr,
            "max_lr": self.max_lr,
            "step_size": self.step_size,
            "mode": self.mode,
            "gamma": self.gamma,
            "scale_mode": self.scale_mode,
        }



import tensorflow as tf
from keras import backend as K


@tf.keras.utils.register_keras_serializable("dlomix")
def adjusted_mean_absolute_error(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """
    Used as an evaluation metric for charge state prediction.

    For two vectors, discard those components that
    are 0 in both vectors and compute the mean
    absolute error for the adjusted vector.
    """
    # Convert y_true and y_pred to float tensors
    y_true = K.cast(y_true, dtype="float32")
    y_pred = K.cast(y_pred, dtype="float32")

    # Create a mask for elements that are not both zero
    mask = K.cast(K.not_equal(y_true + y_pred, 0.0), dtype="float32")

    # Apply mask to both y_true and y_pred
    y_true_adjusted = y_true * mask
    y_pred_adjusted = y_pred * mask

    # Compute the mean absolute error
    absolute_errors = K.abs(y_true_adjusted - y_pred_adjusted)
    sum_absolute_errors = K.sum(absolute_errors)
    count_non_zero = K.sum(mask)

    # Avoid division by zero by adding a small epsilon to the denominator
    epsilon = K.epsilon()
    mean_absolute_error = sum_absolute_errors / (count_non_zero + epsilon)

    return mean_absolute_error


@tf.keras.utils.register_keras_serializable("dlomix")
def adjusted_mean_squared_error(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """
    For two vectors, discard those components that
    are 0 in both vectors and compute the mean
    squared error for the adjusted vector.
    """
    # Convert y_true and y_pred to float tensors
    y_true = K.cast(y_true, dtype="float32")
    y_pred = K.cast(y_pred, dtype="float32")

    # Create a mask for elements that are not both zero
    mask = K.cast(K.not_equal(y_true + y_pred, 0.0), dtype="float32")

    # Apply mask to both y_true and y_pred
    y_true_adjusted = y_true * mask
    y_pred_adjusted = y_pred * mask

    # Compute the mean squared error
    squared_errors = K.square(y_true_adjusted - y_pred_adjusted)
    sum_squared_errors = K.sum(squared_errors)
    count_non_zero = K.sum(mask)

    # Avoid division by zero by adding a small epsilon to the denominator
    epsilon = K.epsilon()
    mean_squared_error = sum_squared_errors / (count_non_zero + epsilon)

    return mean_squared_error


if __name__ == "__main__":
    import numpy as np

    y_true = K.constant([0, 1, 2, 2, 0, 0, 0, 0], dtype="float32")
    y_pred = K.constant([0, 3, 0, 4, 0, 0, 2, 0], dtype="float32")
    y_true, y_pred = K.to_dense(y_true), K.to_dense(y_pred)
    mae = adjusted_mean_absolute_error(y_true, y_pred)
    mse = adjusted_mean_squared_error(y_true, y_pred)
    assert np.isclose(mae, 2.0)
    assert np.isclose(mse, 4.0)
    print(f"Adjusted MAE: {mae:.4f}")
    print(f"Adjusted MSE: {mse:.4f}")



from typing import Callable, Dict, List, Optional, Union

from .dataset import PeptideDataset
from .dataset_config import DatasetConfig
from .dataset_utils import EncodingScheme


class ChargeStateDataset(PeptideDataset):
    """
    A dataset class for handling charge state prediction data.

    Args:
        data_source (Optional[Union[str, List]]): The path or list of paths to the data source file(s).
        val_data_source (Optional[Union[str, List]]): The path or list of paths to the validation data source file(s).
        test_data_source (Optional[Union[str, List]]): The path or list of paths to the test data source file(s).
        data_format (str): The format of the data source file(s). Default is "parquet".
        sequence_column (str): The name of the column containing the peptide sequences. Default is "modified_sequence".
        label_column (str): The name of the column containing the charge state labels. Default is "most_abundant_charge_by_count".
        val_ratio (float): The ratio of validation data to split from the training data. Default is 0.2.
        max_seq_len (Union[int, str]): The maximum length of the peptide sequences. Default is 30.
        dataset_type (str): The type of dataset to use. Default is "tf". Fallback is to TensorFlow dataset tensors.
        batch_size (int): The batch size for training and evaluation. Default is 256.
        shuffle (bool): Whether to shuffle the data. Default is False.
        model_features (Optional[List[str]]): The list of features to use for the model. Default is None.
        dataset_columns_to_keep (Optional[List[str]]): The list of columns to keep in the dataset. Default is None.
        features_to_extract (Optional[List[Union[Callable, str]]]): The list of features to extract from the dataset. Default is None.
        pad (bool): Whether to pad the sequences to the maximum length. Default is True.
        padding_value (str): The value to use for padding. Default is '-'.
        alphabet (Optional[Dict]): The mapping of characters to integers for encoding the sequences. Default is None to trigger learning the alphabet.
        with_termini (bool): Whether to add the N- and C-termini in the sequence column, even if they do not exist. Defaults to True.
        encoding_scheme (Union[str, EncodingScheme]): The encoding scheme to use for encoding the sequences. Default is EncodingScheme.UNMOD.
        processed (bool): Whether the data has been preprocessed. Default is False.
        enable_tf_dataset_cache (bool): Flag to indicate whether to enable TensorFlow Dataset caching (call `.cache()` on the generated TF Datasets).
        disable_cache (bool): Whether to disable Hugging Face datasets caching. Default is False.
        auto_cleanup_cache (bool): Whether to automatically clean up the cache. Default is True.
        num_proc (Optional[int]): Number of processes to use for dataset processing. Use -1 for all available processors, None for single-process mode, or a positive integer. Default is -1.
        batch_processing_size (int): Size of batches for processing. Default is 1000.
        torch_dataloader_kwargs (Optional[Dict]): Additional keyword arguments to pass to PyTorch DataLoader. Default is None.
    """

    def __init__(
        self,
        data_source: Optional[Union[str, List]] = None,
        val_data_source: Optional[Union[str, List]] = None,
        test_data_source: Optional[Union[str, List]] = None,
        data_format: str = "parquet",
        sequence_column: str = "modified_sequence",
        label_column: str = "most_abundant_charge_by_count",
        val_ratio: float = 0.2,
        max_seq_len: Union[int, str] = 30,
        dataset_type: str = "tf",
        batch_size: int = 256,
        shuffle: bool = False,
        model_features: Optional[List[str]] = None,
        dataset_columns_to_keep: Optional[List[str]] = None,
        features_to_extract: Optional[List[Union[Callable, str]]] = None,
        pad: bool = True,
        padding_value: str = "-",
        alphabet: Optional[Dict] = None,
        with_termini: bool = True,
        encoding_scheme: Union[str, EncodingScheme] = EncodingScheme.UNMOD,
        processed: bool = False,
        enable_tf_dataset_cache: bool = False,
        disable_cache: bool = False,
        auto_cleanup_cache: bool = True,
        num_proc: Optional[int] = -1,
        batch_processing_size: int = 1000,
        torch_dataloader_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        config_kwargs = {
            k: v
            for k, v in locals().items()
            if k not in ["self", "__class__", "kwargs"]
        }
        super().__init__(DatasetConfig(**config_kwargs), **kwargs)



import torch
import torch.nn as nn
import torch.nn.functional as F


class DecoderAttentionLayer(nn.Module):
    """
    Decoder attention layer.

    This layer computes attention weights over the time steps for each feature and then
    multiplies the input by these weights. It is assumed that the input tensor has shape
    (batch_size, time_steps, features), and that the provided `time_steps` matches the
    size of the time dimension.

    Parameters
    ----------
    time_steps : int
        Number of time steps in the input data (i.e. the sequence dimension)
    """

    def __init__(self, time_steps):
        super(DecoderAttentionLayer, self).__init__()
        self.time_steps = time_steps
        # This linear layer maps a vector of length time_steps to a vector of length time_steps.
        self.linear = nn.LazyLinear(time_steps)

    def forward(self, x):
        """
        Forward pass of the layer.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch_size, time_steps, features).

        Returns
        -------
        torch.Tensor
            Output tensor of the same shape as the input.
        """
        # Permute the tensor from (batch, time_steps, features) to (batch, features, time_steps)
        x_permuted = x.transpose(1, 2)

        # Apply the linear layer on the last dimension (which has size time_steps)
        # and then apply softmax along that dimension.
        attn = self.linear(x_permuted)
        attn = F.softmax(attn, dim=-1)

        # Permute the attention tensor back to (batch, time_steps, features)
        attn = attn.transpose(1, 2)
        # Multiply the original input with the attention weights elementwise

        out = x * attn
        return out


class AttentionLayer(nn.Module):
    """
    Attention layer.

    This layer computes a weighted average over time steps of the input sequence.
    It first computes an “attention score” for each time step by performing a dot
    product between the input at that time step and a learned weight vector (and optionally
    adding a learned bias per time step). It then applies a tanh nonlinearity (unless
    the `context` flag is True, in which case the tanh result is replaced by another dot
    product with a learned context vector), exponentiates the result, applies an optional mask,
    normalizes the scores to form a probability distribution over time, and then returns
    the weighted sum over the time dimension.

    Parameters
    ----------
    feature_dim : int
        The number of features (i.e. the last dimension of the input tensor).
    seq_len : int
        The fixed length of the time dimension of the input.
    context : bool, optional
        Whether to use a separate context vector. If True, the attention score is computed as
        a dot product with this vector (ignoring the tanh of the weighted sum). Defaults to False.
    bias : bool, optional
        Whether to use a learned bias (with one bias per time step). Defaults to True.
    """

    def __init__(self, feature_dim, seq_len, context=False, bias=True, epsilon=1e-8):
        super(AttentionLayer, self).__init__()
        self.feature_dim = feature_dim
        self.seq_len = seq_len
        self.context = context
        self.bias = bias
        self.epsilon = epsilon

        # Weight vector W of shape (feature_dim,)
        self.W = nn.Parameter(torch.Tensor(feature_dim))

        # Optional bias of shape (seq_len,)
        if bias:
            self.b = nn.Parameter(torch.Tensor(seq_len))
        else:
            self.register_parameter("b", None)

        # Optional context vector u of shape (feature_dim,)
        if context:
            self.u = nn.Parameter(torch.Tensor(feature_dim))
        else:
            self.register_parameter("u", None)

        self.reset_parameters()

    def reset_parameters(self):
        # Initialize the weights using Xavier (Glorot) uniform initialization.
        nn.init.xavier_uniform_(self.W.unsqueeze(0))
        if self.bias:
            nn.init.zeros_(self.b)
        if self.context:
            nn.init.xavier_uniform_(self.u.unsqueeze(0))

    def forward(self, x, mask=None):
        """
        Forward pass for the attention layer.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch_size, seq_len, feature_dim).
        mask : torch.Tensor, optional
            A mask tensor of shape (batch_size, seq_len) with 1 for valid timesteps and 0 for masked timesteps.
            Defaults to None.

        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch_size, feature_dim) which is the weighted sum over time.
        """
        # Compute the attention scores.
        # First, compute the dot product between each time-step input and the weight vector.
        # x: (batch_size, seq_len, feature_dim)
        # self.W: (feature_dim,) -> result is (batch_size, seq_len)
        a = torch.matmul(x, self.W)

        # Add bias if applicable (bias is per time step)
        if self.bias:
            # self.b has shape (seq_len,), and will be broadcast over the batch dimension.
            a = a + self.b

        # Apply tanh nonlinearity.
        a = torch.tanh(a)

        # If context is used, override a with a new dot product with the context vector.
        if self.context:
            # Compute dot product with context vector: result shape (batch_size, seq_len)
            a = torch.matmul(x, self.u)

        # Exponentiate the attention scores.
        a = torch.exp(a)

        # Apply mask if provided.
        if mask is not None:
            # Ensure mask is float so multiplication works as expected.
            a = a * mask.float()

        # Normalize the attention scores over the time dimension.
        # Add a small epsilon to avoid division by zero.
        a_sum = torch.sum(a, dim=1, keepdim=True) + self.epsilon
        a = a / a_sum  # Shape: (batch_size, seq_len)

        # Expand the attention weights for multiplication with x.
        a = a.unsqueeze(-1)  # Now shape: (batch_size, seq_len, 1)

        # Multiply the attention weights with the input and sum over time steps.
        weighted_input = x * a  # (batch_size, seq_len, feature_dim)
        output = torch.sum(weighted_input, dim=1)  # (batch_size, feature_dim)
        return output



import logging

import numpy as np
import tensorflow as tf
import torch

from dlomix.losses.intensity import (
    masked_pearson_correlation_distance,
    masked_spectral_distance,
)
from dlomix.losses.intensity_torch import (
    masked_pearson_correlation_distance as masked_pearson_correlation_distance_torch,
)
from dlomix.losses.intensity_torch import (
    masked_spectral_distance as masked_spectral_distance_torch,
)

logger = logging.getLogger(__name__)


# ------------------ intensity - masked spectral distance ------------------


def test_tf_torch_equivalence_masked_spectral_distance():
    y_true = [[0.1, 0.2, 0.3]]
    y_pred = [list(reversed(y_true[0]))]

    sa_tf = masked_spectral_distance(
        tf.convert_to_tensor(y_true), tf.convert_to_tensor(y_pred)
    )
    sa_torch = masked_spectral_distance_torch(
        torch.tensor(y_true), torch.tensor(y_pred)
    )

    logger.info(
        f"Spectral Angle: for tf: {sa_tf.numpy()} vs for torch: {sa_torch.numpy()}"
    )

    assert np.allclose(
        sa_tf.numpy(), sa_torch.numpy()
    )  # alternatively try np.array_equiv(A,B)


# ------------------ intensity - masked pearson correlation distance ------------------


def test_tf_torch_equivalence_masked_pearson_correlation_distance():
    y_true = [[0.1, 0.2, 0.3]]
    y_pred = [list(reversed(y_true[0]))]

    pc_tf = masked_pearson_correlation_distance(
        tf.convert_to_tensor(y_true), tf.convert_to_tensor(y_pred)
    )
    pc_torch = masked_pearson_correlation_distance_torch(
        torch.tensor(y_true), torch.tensor(y_pred)
    )

    logger.info(
        f"Masked Pearson Correlation Distance: for tf: {pc_tf.numpy()} vs for torch: {pc_torch.numpy()}"
    )

    assert np.allclose(
        pc_tf.numpy(), pc_torch.numpy()
    )  # alternatively try np.array_equiv(A,B)


# add tests for IonMobLoss



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K


@tf.keras.utils.register_keras_serializable("dlomix")
def masked_spectral_distance(y_true, y_pred):
    """
    Calculates the masked spectral distance between true and predicted intensity vectors.
    The masked spectral distance is a metric for comparing the similarity between two intensity vectors.

    Masked, normalized spectral angles between true and pred vectors

    > arccos(1*1 + 0*0) = 0 -> SL = 0 -> high correlation

    > arccos(0*1 + 1*0) = pi/2 -> SL = 1 -> low correlation

    Parameters
    ----------
    y_true : tf.Tensor
        A tensor containing the true values, with shape `(batch_size, num_values)`.
    y_pred : tf.Tensor
        A tensor containing the predicted values, with the same shape as `y_true`.

    Returns
    -------
    tf.Tensor
        A tensor containing the masked spectral distance between `y_true` and `y_pred`.

    """

    # To avoid numerical instability during training on GPUs,
    # we add a fuzzing constant epsilon of 1×10−7 to all vectors
    epsilon = K.epsilon()

    # Masking: we multiply values by (true + 1) because then the peaks that cannot
    # be there (and have value of -1 as explained above) won't be considered
    pred_masked = ((y_true + 1) * y_pred) / (y_true + 1 + epsilon)
    true_masked = ((y_true + 1) * y_true) / (y_true + 1 + epsilon)

    # L2 norm
    pred_norm = K.l2_normalize(true_masked, axis=-1)
    true_norm = K.l2_normalize(pred_masked, axis=-1)

    # Spectral Angle (SA) calculation
    # (from the definition below, it is clear that ions with higher intensities
    #  will always have a higher contribution)
    product = K.sum(pred_norm * true_norm, axis=1)
    arccos = tf.math.acos(product)
    return 2 * arccos / np.pi


@tf.keras.utils.register_keras_serializable("dlomix")
def masked_pearson_correlation_distance(y_true, y_pred):
    """
    Calculates the masked Pearson correlation distance between true and predicted intensity vectors.
    The masked Pearson correlation distance is a metric for comparing the similarity between two intensity vectors,
    taking into account only the non-negative values in the true values tensor (which represent valid peaks).

    Parameters
    ----------
    y_true : tf.Tensor
        A tensor containing the true values, with shape `(batch_size, num_values)`.
    y_pred : tf.Tensor
        A tensor containing the predicted values, with the same shape as `y_true`.

    Returns
    -------
    tf.Tensor
        A tensor containing the masked Pearson correlation distance between `y_true` and `y_pred`.

    """

    epsilon = K.epsilon()

    # Masking: we multiply values by (true + 1) because then the peaks that cannot
    # be there (and have value of -1 as explained above) won't be considered
    pred_masked = ((y_true + 1) * y_pred) / (y_true + 1 + epsilon)
    true_masked = ((y_true + 1) * y_true) / (y_true + 1 + epsilon)

    mx = tf.math.reduce_mean(
        true_masked
    )  # If axis is None all dimensions are reduced, and a tensor with a single element is returned
    my = tf.math.reduce_mean(pred_masked)
    xm, ym = true_masked - mx, pred_masked - my
    r_num = tf.math.reduce_mean(tf.multiply(xm, ym))
    r_den = tf.math.reduce_std(xm) * tf.math.reduce_std(ym)
    return 1 - (r_num / r_den)



import numpy as np
import tensorflow as tf
import torch

from dlomix.eval.chargestate import (
    adjusted_mean_absolute_error,
    adjusted_mean_squared_error,
)
from dlomix.eval.rt_eval import TimeDeltaMetric, timedelta

Y_TRUE = [0, 1, 2, 2, 0, 0, 0, 0]
Y_PRED = [0, 3, 0, 4, 0, 0, 2, 0]


def test_adjusted_mean_absolute_error():
    y_true = tf.constant(Y_TRUE, dtype="float32")
    y_pred = tf.constant(Y_PRED, dtype="float32")
    mae = adjusted_mean_absolute_error(y_true, y_pred)
    assert np.isclose(mae, 2.0)


def test_adjusted_mean_squared_error():
    y_true = tf.constant(Y_TRUE, dtype="float32")
    y_pred = tf.constant(Y_PRED, dtype="float32")
    mse = adjusted_mean_squared_error(y_true, y_pred)
    assert np.isclose(mse, 4.0)


def test_adjusted_mean_absolute_error_torch():
    y_true = torch.tensor(Y_TRUE, dtype=torch.float32)
    y_pred = torch.tensor(Y_PRED, dtype=torch.float32)
    mae = adjusted_mean_absolute_error(y_true, y_pred)
    assert np.isclose(mae, 2.0)


def test_adjusted_mean_squared_error_torch():
    y_true = torch.tensor(Y_TRUE, dtype=torch.float32)
    y_pred = torch.tensor(Y_PRED, dtype=torch.float32)
    mse = adjusted_mean_squared_error(y_true, y_pred)
    assert np.isclose(mse, 4.0)


def test_rt_eval_rf():
    y_true = tf.constant([1.0, 2.0, 3.0, 4.0, 5.0])
    y_pred = tf.constant([1.5, 3.0, 4.5, 6.0, 7.5])
    # abs_error =        [0.5, 1.0, 1.5, 2.0, 2.5]

    metric = TimeDeltaMetric(double_delta=True)
    metric.update_state(y_true, y_pred)
    assert metric.delta == 4.0
    assert metric.result() == 4.0 / 1
    assert timedelta(y_true, y_pred) == 2


def test_rt_eval_torch():
    y_true = torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0])
    y_pred = torch.tensor([1.5, 3.0, 4.5, 6.0, 7.5])
    # abs_error =         [0.5, 1.0, 1.5, 2.0, 2.5]

    assert timedelta(y_true, y_pred) == 2



import logging

import pytest

from dlomix.data.processing import (
    FunctionProcessor,
    SequenceEncodingProcessor,
    SequencePaddingProcessor,
    SequenceParsingProcessor,
    SequencePTMRemovalProcessor,
)

logger = logging.getLogger(__name__)

SEQ_COLUMN = "sequence"

SEQUENCE_UNMODIFIED = "[]-DEL-[]"
PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI = ["D", "E", "L"]

SEQUENCE_MODIFIED = "[]-HC[UNIMOD:4]VD-[]"
PARSED_SEQUENCE_MODIFIED_NO_TERMINI = ["H", "C[UNIMOD:4]", "V", "D"]

SEQUENCE_MODIFIED_WITH_N_MOD = "[UNIMOD:737]-ILC[UNIMOD:4]SIQGFK[UNIMOD:737]D-[]"
PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI = [
    "I",
    "L",
    "C[UNIMOD:4]",
    "S",
    "I",
    "Q",
    "G",
    "F",
    "K[UNIMOD:737]",
    "D",
]


def assert_parsed_data(
    parsed_data, new_sequence_column_value, parsed_sequence_value, n_term, c_term
):
    assert SequenceParsingProcessor.PARSED_COL_NAMES["seq"] in parsed_data.keys()
    assert SequenceParsingProcessor.PARSED_COL_NAMES["n_term"] in parsed_data.keys()
    assert SequenceParsingProcessor.PARSED_COL_NAMES["c_term"] in parsed_data
    assert SEQ_COLUMN in parsed_data.keys()

    assert (
        parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["seq"]]
        == parsed_sequence_value
    )
    assert parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["n_term"]] == n_term
    assert parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["c_term"]] == c_term
    assert parsed_data[SEQ_COLUMN] == new_sequence_column_value


def test_sequence_parsing_processor_unmodified():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=False)
    input_data = {SEQ_COLUMN: SEQUENCE_UNMODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_batched():
    p = SequenceParsingProcessor(
        sequence_column_name=SEQ_COLUMN, with_termini=False, batched=True
    )
    input_data = {SEQ_COLUMN: [SEQUENCE_UNMODIFIED]}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        [PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI],
        [PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI],
        ["[]-"],
        ["-[]"],
    )


def test_sequence_parsing_processor_with_termini():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_UNMODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[]-", *PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_with_modifications():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_MODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[]-", *PARSED_SEQUENCE_MODIFIED_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_MODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_with_modifications_and_nterm_mods():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_MODIFIED_WITH_N_MOD}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[UNIMOD:737]-", *PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI,
        "[UNIMOD:737]-",
        "-[]",
    )


def test_sequence_padding_processor_keep():
    length = 5
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=length)
    input_data = {SEQ_COLUMN: PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI}
    padded = p(input_data)
    logger.info(padded)

    assert len(padded[SEQ_COLUMN]) == length
    assert padded[SEQ_COLUMN] == ["D", "E", "L", 0, 0]
    assert padded[SequencePaddingProcessor.KEEP_COLUMN_NAME]


def test_sequence_padding_processor_drop():
    length = 2
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=length)
    input_data = {SEQ_COLUMN: PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI}
    padded = p(input_data)
    logger.info(padded)

    assert padded[SEQ_COLUMN] == PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI[:length]
    assert not padded[SequencePaddingProcessor.KEEP_COLUMN_NAME]


def test_sequence_encoding_processor_fixed_alphabet(basic_alphabet):
    """Test encoding with fixed alphabet (unknown tokens -> unknown_token_index)."""
    alphabet = basic_alphabet.copy()
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet,
        extend_alphabet=False,
        unknown_token="X",
        unknown_token_index=23,
    )

    # Known sequence
    input_data = {SEQ_COLUMN: ["D", "E", "L"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == [3, 4, 10]  # D=3, E=4, L=10 from alphabet

    # Unknown token should map to unknown_token_index
    # Note: alphabet is a copy from fixture which doesn't have Z
    input_data = {SEQ_COLUMN: ["Z"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == [23]  # Z unknown -> unknown_token_index


def test_sequence_encoding_processor_fixed_alphabet_with_ptm(ptm_alphabet):
    """Test encoding with PTM modifications in fixed alphabet."""
    alphabet = ptm_alphabet.copy()
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet,
        extend_alphabet=False,
    )

    input_data = {SEQ_COLUMN: ["H", "C[UNIMOD:4]", "V", "D"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == [7, 23, 18, 3]  # H=7, C[UNIMOD:4]=23, V=18, D=3


def test_sequence_encoding_processor_extend_alphabet():
    """Test encoding with alphabet extension (learns new tokens)."""
    initial_alphabet = {
        "A": 1,
        "C": 2,
        "D": 3,
        "E": 4,
        "L": 10,
    }

    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=initial_alphabet,
        extend_alphabet=True,
    )

    # First sequence - learns new tokens
    input_data = {SEQ_COLUMN: ["D", "E", "L"]}
    result = p(input_data)
    assert result[SEQ_COLUMN] == [3, 4, 10]

    # Second sequence with unknown token - extends alphabet
    input_data = {SEQ_COLUMN: ["A", "C", "V"]}
    result = p(input_data)
    # V should now be in the alphabet
    assert "V" in p.alphabet
    assert result[SEQ_COLUMN][2] == p.alphabet["V"]


def test_sequence_encoding_processor_extend_alphabet_batched(basic_alphabet):
    """Test alphabet extension in batched mode."""
    alphabet = basic_alphabet.copy()
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet,
        batched=True,
        extend_alphabet=True,
    )

    # Initialize with unknown amino acids
    initial_alphabet = {"A": 1, "C": 2, "D": 3}
    p.alphabet = initial_alphabet.copy()

    input_data = {SEQ_COLUMN: [["A", "C"], ["D", "E"]]}
    result = p(input_data)

    # "E" should be learned and added to alphabet
    assert "E" in p.alphabet
    assert len(result[SEQ_COLUMN]) == 2
    assert len(result[SEQ_COLUMN][0]) == 2
    assert len(result[SEQ_COLUMN][1]) == 2


def test_sequence_encoding_processor_fallback_mode_unmodified(basic_alphabet):
    """Test fallback mode: unknown PTM -> unmodified amino acid."""
    alphabet = basic_alphabet.copy()
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet,
        extend_alphabet=False,
        fallback_unmodified=True,
    )

    # Unknown PTM C[UNIMOD:999] should fallback to C
    input_data = {SEQ_COLUMN: ["C[UNIMOD:999]"]}
    result = p(input_data)

    # C[UNIMOD:999] -> C -> 2
    assert result[SEQ_COLUMN][0] == 2


def test_sequence_encoding_processor_fallback_mode_terminal_mods(basic_alphabet):
    """Test fallback mode with unknown terminal modifications."""
    alphabet = basic_alphabet.copy()
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet,
        extend_alphabet=False,
        fallback_unmodified=True,
    )

    # Unknown N-terminal mod [UNIMOD:999]- should fallback to []-
    input_data = {SEQ_COLUMN: ["[UNIMOD:999]-", "D", "E"]}
    result = p(input_data)

    assert result[SEQ_COLUMN][0] == 21  # []-
    assert result[SEQ_COLUMN][1] == 3  # D
    assert result[SEQ_COLUMN][2] == 4  # E


def test_sequence_encoding_processor_unknown_token_already_in_alphabet():
    """Test that unknown token uses existing index if already in alphabet."""
    alphabet = {
        "A": 1,
        "X": 2,  # X already in alphabet
        "D": 3,
    }

    # Should not raise, but should warn
    with pytest.warns(UserWarning):
        p = SequenceEncodingProcessor(
            sequence_column_name=SEQ_COLUMN,
            alphabet=alphabet,
            unknown_token="X",
            unknown_token_index=1,  # Will be overridden to 2
        )

    # unknown_token_index should be set to existing X index
    assert p.unknown_token_index == 2


def test_sequence_encoding_processor_unknown_token_index_conflict():
    """Test warning when unknown_token_index is already used."""
    alphabet = {
        "A": 1,
        "C": 2,
        "D": 3,
    }

    # Should warn about index conflict
    with pytest.warns(UserWarning):
        p = SequenceEncodingProcessor(
            sequence_column_name=SEQ_COLUMN,
            alphabet=alphabet,
            unknown_token_index=2,  # 2 is already used by C
        )

    # Index should be reassigned
    assert p.unknown_token_index != 2


def test_sequence_encoding_processor_single_vs_batched_consistency(basic_alphabet):
    """Test that single and batched modes produce identical results."""
    alphabet = basic_alphabet.copy()
    p_single = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet.copy(),
        batched=False,
    )
    p_batch = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet=alphabet.copy(),
        batched=True,
    )

    sequences = [["D", "E", "L"], ["A", "C"], ["K", "L", "M"]]

    # Process individually
    results_single = [p_single({SEQ_COLUMN: seq})[SEQ_COLUMN] for seq in sequences]

    # Process in batch
    result_batch = p_batch({SEQ_COLUMN: sequences})[SEQ_COLUMN]

    assert results_single == result_batch


# ============================================================================
# SEQUENCE PTM REMOVAL PROCESSOR TESTS
# ============================================================================


def test_sequence_ptm_removal_processor_basic():
    """Test basic PTM removal from sequence."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN)

    input_data = {SEQ_COLUMN: ["[]-", "C[UNIMOD:4]", "V", "D", "-[]"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["[]-", "C", "V", "D", "-[]"]


def test_sequence_ptm_removal_processor_multiple_ptms():
    """Test removal of multiple PTMs from same sequence."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN)

    input_data = {
        SEQ_COLUMN: ["[UNIMOD:737]-", "C[UNIMOD:4]", "K[UNIMOD:737]", "S", "-[]"]
    }
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["[UNIMOD:737]-", "C", "K", "S", "-[]"]


def test_sequence_ptm_removal_processor_no_ptms():
    """Test sequence without PTMs is unchanged."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN)

    input_data = {SEQ_COLUMN: ["[]-", "D", "E", "L", "-[]"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["[]-", "D", "E", "L", "-[]"]


def test_sequence_ptm_removal_processor_batched():
    """Test PTM removal in batched mode."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN, batched=True)

    input_data = {
        SEQ_COLUMN: [
            ["[]-", "C[UNIMOD:4]", "V", "-[]"],
            ["[]-", "K[UNIMOD:737]", "D", "-[]"],
        ]
    }
    result = p(input_data)

    assert result[SEQ_COLUMN][0] == ["[]-", "C", "V", "-[]"]
    assert result[SEQ_COLUMN][1] == ["[]-", "K", "D", "-[]"]


def test_sequence_ptm_removal_processor_preserves_terminals():
    """Test that terminal modifications are preserved."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN)

    input_data = {SEQ_COLUMN: ["[UNIMOD:737]-", "C[UNIMOD:4]", "-[UNIMOD:1]"]}
    result = p(input_data)

    # Terminals should be unchanged
    assert result[SEQ_COLUMN][0] == "[UNIMOD:737]-"
    assert result[SEQ_COLUMN][-1] == "-[UNIMOD:1]"


def test_sequence_ptm_removal_processor_non_list_input_raises_error():
    """Test that non-list input raises ValueError."""
    p = SequencePTMRemovalProcessor(sequence_column_name=SEQ_COLUMN)

    # String input should raise error
    input_data = {SEQ_COLUMN: "C[UNIMOD:4]VD"}

    with pytest.raises(ValueError, match="Sequence must be a list"):
        p(input_data)


# ============================================================================
# FUNCTION PROCESSOR TESTS
# ============================================================================


def test_function_processor_basic(sample_custom_function):
    """Test basic function processor application."""
    p = FunctionProcessor(function=sample_custom_function)

    input_data = {SEQ_COLUMN: ["D", "E", "L"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["D", "E", "L", "D", "E", "L"]


def test_function_processor_with_kwargs(sample_custom_function_with_kwargs):
    """Test function processor with keyword arguments."""
    p = FunctionProcessor(function=sample_custom_function_with_kwargs)

    input_data = {"feature": 5.0}
    result = p(input_data, scale_factor=2.0)

    assert result["feature"] == 10.0


def test_function_processor_lambda():
    """Test function processor with lambda function."""
    p = FunctionProcessor(
        function=lambda data, **kwargs: {"sequence": data.get("sequence", []) * 2}
    )

    input_data = {"sequence": [1, 2, 3]}
    result = p(input_data)

    assert result["sequence"] == [1, 2, 3, 1, 2, 3]


# ============================================================================
# EDGE CASE TESTS
# ============================================================================


def test_sequence_padding_processor_exact_length():
    """Test padding when sequence length equals max_length."""
    length = 3
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=length)

    input_data = {SEQ_COLUMN: ["D", "E", "L"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["D", "E", "L"]
    assert result[SequencePaddingProcessor.KEEP_COLUMN_NAME] is True


def test_sequence_padding_processor_custom_padding_index():
    """Test padding with custom padding index."""
    p = SequencePaddingProcessor(
        sequence_column_name=SEQ_COLUMN,
        max_length=5,
        padding_index=-1,
    )

    input_data = {SEQ_COLUMN: ["D", "E", "L"]}
    result = p(input_data)

    assert result[SEQ_COLUMN] == ["D", "E", "L", -1, -1]


def test_sequence_padding_processor_empty_sequence():
    """Test padding with empty sequence."""
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=3)

    input_data = {SEQ_COLUMN: []}
    result = p(input_data)

    assert result[SEQ_COLUMN] == [0, 0, 0]
    assert result[SequencePaddingProcessor.KEEP_COLUMN_NAME] is True


def test_sequence_padding_processor_long_sequence():
    """Test truncation of very long sequence."""
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=5)

    long_sequence = list(range(10))
    input_data = {SEQ_COLUMN: long_sequence}
    result = p(input_data)

    assert len(result[SEQ_COLUMN]) == 5
    assert result[SEQ_COLUMN] == long_sequence[:5]
    assert result[SequencePaddingProcessor.KEEP_COLUMN_NAME] is False


def test_sequence_parsing_processor_invalid_format():
    """Test error on invalid ProForma format (4+ parts)."""
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN)

    input_data = {SEQ_COLUMN: "[UNIMOD:1]-SEQ-[UNIMOD:2]-EXTRA"}

    with pytest.raises(ValueError, match="Invalid sequence format"):
        p(input_data)


def test_sequence_encoding_processor_empty_alphabet_fixed_mode():
    """Test that fixed mode works with empty alphabet + unknown tokens."""
    p = SequenceEncodingProcessor(
        sequence_column_name=SEQ_COLUMN,
        alphabet={},  # Empty alphabet (will get unknown token added)
        extend_alphabet=False,
        unknown_token="X",
        unknown_token_index=1,
    )

    # Alphabet was empty, but unknown token X was added with index 1
    assert "X" in p.alphabet
    assert p.unknown_token_index == 1

    input_data = {SEQ_COLUMN: ["A", "C", "D"]}
    result = p(input_data)

    # All should map to unknown_token_index since they're not in alphabet
    assert result[SEQ_COLUMN] == [1, 1, 1]



from ..config import _BACKEND, PYTORCH_BACKEND, TENSORFLOW_BACKEND

__all__ = []

if _BACKEND in TENSORFLOW_BACKEND:
    from .base import RetentionTimePredictor
    from .chargestate import ChargeStatePredictor
    from .detectability import DetectabilityModel
    from .prosit import PrositIntensityPredictor, PrositRetentionTimePredictor

    __all__.append("RetentionTimePredictor")


elif _BACKEND in PYTORCH_BACKEND:
    from .chargestate_torch import ChargeStatePredictor
    from .detectability_torch import DetectabilityModel
    from .ionmob_torch import Ionmob
    from .prosit_torch import PrositIntensityPredictor, PrositRetentionTimePredictor

    __all__.append("Ionmob")

__all__.extend(
    [
        "ChargeStatePredictor",
        "PrositRetentionTimePredictor",
        "PrositIntensityPredictor",
        "DetectabilityModel",
        "ChargeStatePredictor",
    ]
)



import logging
import time
from os.path import join
from shutil import rmtree

from datasets import Dataset, DatasetDict, load_dataset

from dlomix.data import (
    FragmentIonIntensityDataset,
    RetentionTimeDataset,
    load_processed_dataset,
)
from dlomix.data.dataset_utils import EncodingScheme

logger = logging.getLogger(__name__)

RT_HUB_DATASET_NAME = "Wilhelmlab/prospect-ptms-irt"


def test_empty_rtdataset():
    rtdataset = RetentionTimeDataset()
    assert rtdataset.hf_dataset is None
    assert rtdataset._empty_dataset_mode is True


def test_num_proc_minus_one_uses_available_processors(monkeypatch):
    monkeypatch.setattr("dlomix.data.dataset.get_num_processors", lambda: 6)

    dataset = RetentionTimeDataset(num_proc=-1)

    assert dataset._num_proc == 6


def test_num_proc_none_forces_single_process(monkeypatch):
    monkeypatch.setattr("dlomix.data.dataset.get_num_processors", lambda: 6)

    dataset = RetentionTimeDataset(num_proc=None)

    assert dataset._num_proc is None


def test_num_proc_user_value_is_capped_to_available(monkeypatch):
    monkeypatch.setattr("dlomix.data.dataset.get_num_processors", lambda: 6)

    dataset = RetentionTimeDataset(num_proc=10)

    assert dataset._num_proc == 6


def test_parquet_rtdataset(download_path_for_assets):
    rtdataset = RetentionTimeDataset(
        data_source=join(download_path_for_assets, "file_1.parquet"),
        sequence_column="modified_sequence",
        label_column="indexed_retention_time",
    )
    assert rtdataset.hf_dataset is not None
    assert rtdataset._empty_dataset_mode is False
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0] in list(
        rtdataset.hf_dataset.keys()
    )
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[1] in list(
        rtdataset.hf_dataset.keys()
    )
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[2] not in list(
        rtdataset.hf_dataset.keys()
    )
    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0]].num_rows > 0
    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[1]].num_rows > 0


def test_rtdataset_inmemory(download_path_for_assets):
    hf_dataset = load_dataset(
        "parquet",
        data_files=join(download_path_for_assets, "file_1.parquet"),
        split="train",
    )

    rtdataset = RetentionTimeDataset(
        data_source=hf_dataset,
        data_format="hf",
        sequence_column="modified_sequence",
        label_column="indexed_retention_time",
    )
    assert rtdataset.hf_dataset is not None
    assert rtdataset._empty_dataset_mode is False
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0] in list(
        rtdataset.hf_dataset.keys()
    )

    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0]].num_rows > 0


def test_rtdataset_hub():
    rtdataset = RetentionTimeDataset(
        data_source=RT_HUB_DATASET_NAME,
        data_format="hub",
        sequence_column="modified_sequence",
        label_column="indexed_retention_time",
        name="holdout",
    )
    logger.info(rtdataset)
    assert rtdataset.hf_dataset is not None
    assert rtdataset._empty_dataset_mode is False

    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[2] in list(
        rtdataset.hf_dataset.keys()
    )

    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[2]].num_rows > 0


def test_csv_rtdataset(download_path_for_assets):
    rtdataset = RetentionTimeDataset(
        data_source=join(download_path_for_assets, "file_2.csv"),
        data_format="csv",
        sequence_column="sequence",
        label_column="irt",
        val_ratio=0.2,
    )

    assert rtdataset.hf_dataset is not None
    assert rtdataset._empty_dataset_mode is False
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0] in list(
        rtdataset.hf_dataset.keys()
    )
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[1] in list(
        rtdataset.hf_dataset.keys()
    )
    assert RetentionTimeDataset.DEFAULT_SPLIT_NAMES[2] not in list(
        rtdataset.hf_dataset.keys()
    )
    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[0]].num_rows > 0
    assert rtdataset[RetentionTimeDataset.DEFAULT_SPLIT_NAMES[1]].num_rows > 0


def test_empty_intensitydataset():
    intensity_dataset = FragmentIonIntensityDataset()
    assert intensity_dataset.hf_dataset is None
    assert intensity_dataset._empty_dataset_mode is True


def test_parquet_intensitydataset(download_path_for_assets):
    filepath = join(download_path_for_assets, "file_3.parquet")
    intensity_dataset = FragmentIonIntensityDataset(
        data_format="parquet",
        data_source=filepath,
        sequence_column="sequence",
        label_column="intensities",
        model_features=["precursor_charge_onehot", "collision_energy_aligned_normed"],
    )

    assert intensity_dataset.hf_dataset is not None
    assert intensity_dataset._empty_dataset_mode is False
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0] in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[1] in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[2] not in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert (
        intensity_dataset[FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0]].num_rows
        > 0
    )
    assert (
        intensity_dataset[FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[1]].num_rows
        > 0
    )


def test_csv_intensitydataset(download_path_for_assets):
    filepath = join(download_path_for_assets, "file_4.csv")
    intensity_dataset = FragmentIonIntensityDataset(
        data_format="csv",
        data_source=filepath,
        sequence_column="sequence",
        label_column="intensities",
    )

    assert intensity_dataset.hf_dataset is not None
    assert intensity_dataset._empty_dataset_mode is False
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0] in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[1] in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[2] not in list(
        intensity_dataset.hf_dataset.keys()
    )
    assert (
        intensity_dataset[FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0]].num_rows
        > 0
    )
    assert (
        intensity_dataset[FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[1]].num_rows
        > 0
    )


def test_nested_model_features(raw_generic_nested_data):
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    intensity_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        model_features=["nested_feature"],
    )

    assert intensity_dataset.hf_dataset is not None
    assert intensity_dataset._empty_dataset_mode is False

    example = iter(intensity_dataset.tensor_train_data).next()
    assert example[0]["nested_feature"].shape == [2, 1, 2]


def test_save_dataset(raw_generic_nested_data):
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    intensity_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        model_features=["nested_feature"],
    )

    save_path = "./.test_dataset_2"
    attributes = intensity_dataset.__dict__
    logger.info("Dataset attributes before saving: {}".format(attributes))

    intensity_dataset.save_to_disk(save_path, overwrite=True)
    rmtree(save_path)


def test_load_dataset(download_path_for_assets):
    rtdataset = RetentionTimeDataset(
        data_source=join(download_path_for_assets, "file_2.csv"),
        data_format="csv",
        sequence_column="sequence",
        label_column="irt",
        val_ratio=0.2,
    )

    save_path = "./.test_dataset_1"
    rtdataset.save_to_disk(save_path, overwrite=True)
    splits = rtdataset._data_files_available_splits
    config = rtdataset._config

    load_time_threshold = 0.05  # 50ms

    start_time = time.time()
    loaded_dataset = load_processed_dataset(save_path)
    load_duration = time.time() - start_time
    logger.info("Loaded the dataset in {} seconds".format(load_duration))

    logger.info("Original datasets config: {}".format(rtdataset._config))
    logger.info("Loaded datasets config: {}".format(loaded_dataset._config))

    # Assert the load time is below the threshold
    assert (
        load_duration < load_time_threshold
    ), f"Load time exceeded: {load_duration:.3f}s"
    assert loaded_dataset.processed is True

    assert loaded_dataset._data_files_available_splits == splits
    assert loaded_dataset.hf_dataset is not None
    assert loaded_dataset._config == config, f"{loaded_dataset._config} != {config}"
    rmtree(save_path)


def test_no_split_datasetDict_hf_inmemory(raw_generic_nested_data):
    hfdata = Dataset.from_dict(raw_generic_nested_data)
    hf_dataset = DatasetDict({"train": hfdata})

    intensity_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hf_dataset,
        sequence_column="seq",
        label_column="label",
    )

    assert intensity_dataset.hf_dataset is not None
    assert intensity_dataset._empty_dataset_mode is False
    assert FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0] in list(
        intensity_dataset.hf_dataset.keys()
    )

    assert (
        len(
            intensity_dataset.hf_dataset[
                FragmentIonIntensityDataset.DEFAULT_SPLIT_NAMES[0]
            ]
        )
        == 2
    )

    # test learning alphabet for train/val and then using it for test with fallback


def _make_rt_split_data(seqs):
    return Dataset.from_dict(
        {
            "modified_sequence": seqs,
            "indexed_retention_time": [0.1 + i for i in range(len(seqs))],
        }
    )


def test_encoding_learning_forces_single_process(monkeypatch):
    # Keep split insertion order as train -> test -> val to capture ordering assumptions.
    hf_dataset = DatasetDict(
        {
            "train": _make_rt_split_data(["[]-AC-[]"]),
            "test": _make_rt_split_data(["[]-C[UNIMOD:4]A-[]"]),
            "val": _make_rt_split_data(["[]-C[UNIMOD:4]A-[]"]),
        }
    )

    calls = []
    original_map = Dataset.map

    def map_spy(self, function, *args, **kwargs):
        calls.append((kwargs.get("desc"), kwargs.get("num_proc")))
        return original_map(self, function, *args, **kwargs)

    monkeypatch.setattr(Dataset, "map", map_spy)

    RetentionTimeDataset(
        data_format="hf",
        data_source=hf_dataset,
        sequence_column="modified_sequence",
        label_column="indexed_retention_time",
        encoding_scheme=EncodingScheme.NAIVE_MODS,
        alphabet=None,
        num_proc=2,
        max_seq_len=8,
    )

    encoding_calls = [c for c in calls if c[0] == "Mapping SequenceEncodingProcessor"]
    assert len(encoding_calls) == 3

    # Encoding is deterministic train -> val -> test.
    # train/val must force single-process learning, test keeps configured num_proc.
    assert encoding_calls[0][1] is None
    assert encoding_calls[1][1] is None
    assert encoding_calls[2][1] == 2


def test_val_tokens_available_to_test_even_with_nonstandard_split_order():
    # Token appears in val and test, but not train. If test is encoded before val,
    # fallback may be used incorrectly instead of learned token encoding.
    hf_dataset = DatasetDict(
        {
            "train": _make_rt_split_data(["[]-AC-[]"]),
            "test": _make_rt_split_data(["[]-C[UNIMOD:4]A-[]"]),
            "val": _make_rt_split_data(["[]-C[UNIMOD:4]A-[]"]),
        }
    )

    dataset = RetentionTimeDataset(
        data_format="hf",
        data_source=hf_dataset,
        sequence_column="modified_sequence",
        label_column="indexed_retention_time",
        encoding_scheme=EncodingScheme.NAIVE_MODS,
        alphabet=None,
        num_proc=2,
        max_seq_len=8,
    )

    assert "C[UNIMOD:4]" in dataset.extended_alphabet

    test_encoded = dataset.hf_dataset["test"][0]["modified_sequence"]
    learned_token_index = dataset.extended_alphabet["C[UNIMOD:4]"]

    # with_termini=True means sequence starts with []- at index 0.
    assert test_encoded[1] == learned_token_index


def test_shuffle_parameter(raw_generic_nested_data):
    """Test that shuffle parameter works for both TensorFlow and PyTorch datasets."""
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    # Test with shuffle=True for TensorFlow
    tf_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        dataset_type="tf",
        shuffle=True,
        batch_size=1,
    )

    # Test with shuffle=True for PyTorch
    pt_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        dataset_type="pt",
        shuffle=True,
        batch_size=1,
    )

    # Verify datasets are created successfully
    assert tf_dataset.shuffle is True
    assert pt_dataset.shuffle is True
    assert tf_dataset.tensor_train_data is not None
    assert pt_dataset.tensor_train_data is not None


def test_torch_dataloader_kwargs(raw_generic_nested_data):
    """Test that additional PyTorch DataLoader kwargs are properly passed through."""
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        dataset_type="pt",
        batch_size=1,
        torch_dataloader_kwargs={
            "drop_last": True,
            "pin_memory": False,
            "num_workers": 0,  # Use 0 to avoid multiprocessing issues in tests
        },
    )

    # Get the DataLoader
    dataloader = dataset.tensor_train_data

    # Verify that torch_dataloader_kwargs were applied
    assert dataloader.drop_last is True
    assert dataloader.pin_memory is False
    assert dataloader.num_workers == 0
    assert dataset.torch_dataloader_kwargs is not None


def test_tf_tensor_dataset_string_label(raw_generic_nested_data):
    """Test that TensorFlow TensorDataset is created properly."""
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        dataset_type="tf",
        batch_size=1,
    )

    # Get the TensorFlow dataset
    tf_dataset = dataset.tensor_train_data

    # Verify that the TensorFlow dataset is created successfully
    assert tf_dataset is not None
    for batch in tf_dataset.take(1):
        features, labels = batch
        assert features is not None
        assert labels is not None


def test_tf_tensor_dataset_singelton_list_label(raw_generic_nested_data):
    """Test that TensorFlow TensorDataset is created properly."""
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column=["label"],
        dataset_type="tf",
        batch_size=1,
    )

    # Get the TensorFlow dataset
    tf_dataset = dataset.tensor_train_data

    # Verify that the TensorFlow dataset is created successfully
    assert tf_dataset is not None
    for batch in tf_dataset.take(1):
        features, labels = batch
        assert features is not None
        assert labels is not None


def test_tf_tensor_dataset_list_multi_label(raw_generic_nested_data):
    """Test that TensorFlow TensorDataset is created properly."""
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column=["label", "label2"],
        dataset_type="tf",
        batch_size=1,
    )

    # Get the TensorFlow dataset
    tf_dataset = dataset.tensor_train_data

    # Verify that the TensorFlow dataset is created successfully
    assert tf_dataset is not None
    for batch in tf_dataset.take(1):
        features, labels = batch
        assert features is not None
        assert labels is not None



import torch
import torch.nn as nn


class MaskedIonmobLoss(nn.Module):
    def __init__(self, use_mse: bool = True):
        """
        This loss can be used for the training of an ion-mobility predictor that
        outputs both expected ion mean and ion standard deviation.

        Args:
            use_mse: if false, MAE will be used instead of MSE.
        """
        super(MaskedIonmobLoss, self).__init__()
        self.loss_function = nn.MSELoss() if use_mse else nn.L1Loss()

    def forward(self, outputs, targets):
        """
        Computes loss for Ionmob model with masked CCS standard deviation loss.

        Args:
            outputs: Tuple[CCS, CCS_STD]
            targets: Tuple[CCS, CCS_STD]

        Returns:
            Combined loss of predicted CCS and masked CCS STD.
        """
        ccs_output, ccs_std_output = outputs
        target_ccs, target_ccs_std = targets

        # Ensure all tensors are on the same device
        device = ccs_output.device
        target_ccs = target_ccs.to(device)
        target_ccs_std = target_ccs_std.to(device)

        # Compute MSE/MAE loss for CCS
        loss_ccs = self.loss_function(ccs_output, target_ccs)

        # Masked MSE/MAE loss for CCS-STD (ignore -1 values)
        mask = target_ccs_std != -1

        if mask.any():  # Ensure at least one valid target exists
            loss_ccs_std = self.loss_function(
                ccs_std_output[mask], target_ccs_std[mask]
            )
        else:
            loss_ccs_std = torch.tensor(0.0, device=device)

        loss_total = loss_ccs + loss_ccs_std
        return loss_total
